"""RouteDEM for exposing the natcap.invest's routing package to UI."""
import os
import logging

from osgeo import gdal
import pygeoprocessing
import pygeoprocessing.routing
import taskgraph
import numpy

from . import utils
from . import spec_utils
from .spec_utils import u
from . import validation
from . import MODEL_METADATA

LOGGER = logging.getLogger(__name__)

INVALID_BAND_INDEX_MSG = _('Must be between 1 and {maximum}')

ARGS_SPEC = {
    "model_name": MODEL_METADATA["routedem"].model_title,
    "pyname": MODEL_METADATA["routedem"].pyname,
    "userguide_html": MODEL_METADATA["routedem"].userguide,
    "args": {
        "workspace_dir": spec_utils.WORKSPACE,
        "results_suffix": spec_utils.SUFFIX,
        "n_workers": spec_utils.N_WORKERS,
        "dem_path": spec_utils.DEM,
        "dem_band_index": {
            "type": "number",
            "expression": "value >= 1",
            "units": u.none,
            "required": False,
            "about": _(
                "Index of the raster band to use, for multi-band rasters."),
            "name": _("band index")
        },
        "algorithm": {
            "type": "option_string",
            "options": {
                "D8": {
                    "display_name": "D8",
                    "description": _(
                        "All water on a pixel flows into the most downhill of "
                        "its 8 surrounding pixels")},
                "MFD": {
                    "display_name": "MFD",
                    "description": _(
                        "Flow off a pixel is modeled fractionally so that "
                        "water is split among multiple downstream pixels")}
            },
            "about": _("The routing algorithm to use."),
            "name": _("routing algorithm")
        },
        "calculate_flow_direction": {
            "type": "boolean",
            "required": False,
            "about": _("Calculate flow direction from the provided DEM."),
            "name": _("calculate flow direction")
        },
        "calculate_flow_accumulation": {
            "type": "boolean",
            "required": False,
            "about": _(
                "Calculate flow accumulation from the flow direction output."),
            "name": _("calculate flow accumulation")
        },
        "calculate_stream_threshold": {
            "type": "boolean",
            "required": False,
            "about": _(
                "Calculate streams from the flow accumulation output. "),
            "name": _("calculate streams")
        },
        "threshold_flow_accumulation": {
            **spec_utils.THRESHOLD_FLOW_ACCUMULATION,
            "required": "calculate_stream_threshold",
            "about": _(
                f"{spec_utils.THRESHOLD_FLOW_ACCUMULATION['about']} "
                "Required if Calculate Streams is selected.")
        },
        "calculate_downstream_distance": {
            "type": "boolean",
            "required": False,
            "about": _(
                "Calculate flow distance from each pixel to a stream as "
                "defined in the Calculate Streams output."),
            "name": _("calculate distance to stream")
        },
        "calculate_slope": {
            "type": "boolean",
            "required": False,
            "about": _("Calculate percent slope from the provided DEM."),
            "name": _("calculate slope")
        }
    }
}


# replace %s with file suffix
_TARGET_FILLED_PITS_FILED_PATTERN = 'filled%s.tif'
_TARGET_SLOPE_FILE_PATTERN = 'slope%s.tif'
_TARGET_FLOW_DIRECTION_FILE_PATTERN = 'flow_direction%s.tif'
_FLOW_ACCUMULATION_FILE_PATTERN = 'flow_accumulation%s.tif'
_STREAM_MASK_FILE_PATTERN = 'stream_mask%s.tif'
_DOWNSTREAM_DISTANCE_FILE_PATTERN = 'downstream_distance%s.tif'

_ROUTING_FUNCS = {
    'D8': {
        'flow_accumulation': pygeoprocessing.routing.flow_accumulation_d8,
        'flow_direction': pygeoprocessing.routing.flow_dir_d8,
        'threshold_flow': None,  # Defined in source code as _threshold_flow
        'distance_to_channel': pygeoprocessing.routing.distance_to_channel_d8,
    },
    'MFD': {
        'flow_accumulation': pygeoprocessing.routing.flow_accumulation_mfd,
        'flow_direction': pygeoprocessing.routing.flow_dir_mfd,
        'threshold_flow': pygeoprocessing.routing.extract_streams_mfd,
        'distance_to_channel': pygeoprocessing.routing.distance_to_channel_mfd,
    }
}


def _threshold_flow(flow_accum_pixels, threshold, in_nodata, out_nodata):
    """Raster_calculator local_op to threshold D8 stream flow.

    Args:
        flow_accum_pixels (numpy.ndarray): Array representing the number of
            pixels upstream of a given pixel.
        threshold (int or float): The threshold above which we have a stream.
        in_nodata (int or float): The nodata value of the flow accumulation
            raster.
        out_nodata (int): The nodata value of the target stream mask raster.

    Returns:
        A numpy.ndarray (dtype is numpy.uint8) with pixel values of 1 where
        flow accumulation > threshold, 0 where flow accumulation < threshold
        and out_nodata where flow accumulation is equal to in_nodata.

    """
    out_matrix = numpy.empty(flow_accum_pixels.shape, dtype=numpy.uint8)
    out_matrix[:] = out_nodata
    stream_mask = (flow_accum_pixels > threshold)

    valid_mask = slice(None)
    if in_nodata is not None:
        valid_mask = ~utils.array_equals_nodata(flow_accum_pixels, in_nodata)

    out_matrix[valid_mask & stream_mask] = 1
    out_matrix[valid_mask & ~stream_mask] = 0
    return out_matrix


def execute(args):
    """RouteDEM: Hydrological routing.

    This model exposes the pygeoprocessing D8 and Multiple Flow Direction
    routing functionality as an InVEST model.

    This tool will always fill pits on the input DEM.

    Args:
        args['workspace_dir'] (string): output directory for intermediate,
            temporary, and final files
        args['results_suffix'] (string): (optional) string to append to any
            output file names
        args['dem_path'] (string): path to a digital elevation raster
        args['dem_band_index'] (int): Optional. The band index to operate on.
            If not provided, band index 1 is assumed.
        args['algorithm'] (string): The routing algorithm to use.  Must be
            one of 'D8' or 'MFD' (case-insensitive). Required when calculating
            flow direction, flow accumulation, stream threshold, and downstream
            distance.
        args['calculate_flow_direction'] (bool): If True, model will calculate
            flow direction for the filled DEM.
        args['calculate_flow_accumulation'] (bool): If True, model will
            calculate a flow accumulation raster. Only applies when
            args['calculate_flow_direction'] is True.
        args['calculate_stream_threshold'] (bool): if True, model will
            calculate a stream classification layer by thresholding flow
            accumulation to the provided value in
            ``args['threshold_flow_accumulation']``.  Only applies when
            args['calculate_flow_accumulation'] and
            args['calculate_flow_direction'] are True.
        args['threshold_flow_accumulation'] (int): The number of upstream
            cells that must flow into a cell before it's classified as a
            stream.
        args['calculate_downstream_distance'] (bool): If True, and a stream
            threshold is calculated, model will calculate a downstream
            distance raster in units of pixels. Only applies when
            args['calculate_flow_accumulation'],
            args['calculate_flow_direction'], and
            args['calculate_stream_threshold'] are all True.
        args['calculate_slope'] (bool):  If True, model will calculate a
            slope raster from the DEM.
        args['n_workers'] (int): The ``n_workers`` parameter to pass to
            the task graph.  The default is ``-1`` if not provided.

    Returns:
        ``None``
    """
    file_suffix = utils.make_suffix_string(args, 'results_suffix')
    task_cache_dir = os.path.join(args['workspace_dir'], '_taskgraph_working_dir')
    utils.make_directories([args['workspace_dir'], task_cache_dir])

    if ('calculate_flow_direction' in args and
            bool(args['calculate_flow_direction'])):
        # All routing functions depend on this one task.
        # Check the algorithm early so we can fail quickly, but only if we're
        # doing some sort of hydological routing
        algorithm = args['algorithm'].upper()
        try:
            routing_funcs = _ROUTING_FUNCS[algorithm]
        except KeyError:
            raise RuntimeError(
                'Invalid algorithm specified (%s). Must be one of %s' % (
                    args['algorithm'], ', '.join(sorted(_ROUTING_FUNCS.keys()))))

    if 'dem_band_index' in args and args['dem_band_index'] not in (None, ''):
        band_index = int(args['dem_band_index'])
    else:
        band_index = 1
    LOGGER.info('Using DEM band index %s', band_index)

    dem_raster_path_band = (args['dem_path'], band_index)

    try:
        n_workers = int(args['n_workers'])
    except (KeyError, ValueError, TypeError):
        # KeyError when n_workers is not present in args
        # ValueError when n_workers is an empty string.
        # TypeError when n_workers is None.
        n_workers = -1  # Synchronous mode.

    graph = taskgraph.TaskGraph(task_cache_dir, n_workers=n_workers)

    # Calculate slope.  This is intentionally on the original DEM, not
    # on the pitfilled DEM.  If the user really wants the slop of the filled
    # DEM, they can pass it back through RouteDEM.
    if 'calculate_slope' in args and bool(args['calculate_slope']):
        target_slope_path = os.path.join(
            args['workspace_dir'], _TARGET_SLOPE_FILE_PATTERN % file_suffix)
        graph.add_task(
            pygeoprocessing.calculate_slope,
            args=(dem_raster_path_band,
                  target_slope_path),
            task_name='calculate_slope',
            target_path_list=[target_slope_path])

    dem_filled_pits_path = os.path.join(
        args['workspace_dir'],
        _TARGET_FILLED_PITS_FILED_PATTERN % file_suffix)
    filled_pits_task = graph.add_task(
        pygeoprocessing.routing.fill_pits,
        args=(dem_raster_path_band,
              dem_filled_pits_path,
              args['workspace_dir']),
        task_name='fill_pits',
        target_path_list=[dem_filled_pits_path])

    if ('calculate_flow_direction' in args and
            bool(args['calculate_flow_direction'])):
        LOGGER.info("calculating flow direction")
        flow_dir_path = os.path.join(
            args['workspace_dir'],
            _TARGET_FLOW_DIRECTION_FILE_PATTERN % file_suffix)
        flow_direction_task = graph.add_task(
            routing_funcs['flow_direction'],
            args=((dem_filled_pits_path, 1),  # PGP>1.9.0 creates 1-band fills
                  flow_dir_path,
                  args['workspace_dir']),
            target_path_list=[flow_dir_path],
            dependent_task_list=[filled_pits_task],
            task_name='flow_dir_%s' % algorithm)

        if ('calculate_flow_accumulation' in args and
                bool(args['calculate_flow_accumulation'])):
            LOGGER.info("calculating flow accumulation")
            flow_accumulation_path = os.path.join(
                args['workspace_dir'],
                _FLOW_ACCUMULATION_FILE_PATTERN % file_suffix)
            flow_accum_task = graph.add_task(
                routing_funcs['flow_accumulation'],
                args=((flow_dir_path, 1),
                      flow_accumulation_path
                ),
                target_path_list=[flow_accumulation_path],
                task_name='flow_accumulation_%s' % algorithm,
                dependent_task_list=[flow_direction_task])

            if ('calculate_stream_threshold' in args and
                    bool(args['calculate_stream_threshold'])):
                stream_mask_path = os.path.join(
                        args['workspace_dir'],
                    _STREAM_MASK_FILE_PATTERN % file_suffix)
                if algorithm == 'D8':
                    flow_accum_task.join()
                    flow_accum_info = pygeoprocessing.get_raster_info(
                        flow_accumulation_path)
                    stream_threshold_task = graph.add_task(
                        pygeoprocessing.raster_calculator,
                        args=(((flow_accumulation_path, 1),
                               (float(args['threshold_flow_accumulation']), 'raw'),
                               (flow_accum_info['nodata'][0], 'raw'),
                               (255, 'raw')),
                              _threshold_flow,
                              stream_mask_path,
                              gdal.GDT_Byte,
                              255),
                        target_path_list=[stream_mask_path],
                        task_name='stream_thresholding_D8',
                        dependent_task_list=[flow_accum_task])
                else:  # MFD
                    stream_threshold_task = graph.add_task(
                        routing_funcs['threshold_flow'],
                        args=((flow_accumulation_path, 1),
                              (flow_dir_path, 1),
                              float(args['threshold_flow_accumulation']),
                              stream_mask_path),
                        target_path_list=[stream_mask_path],
                        task_name=['stream_extraction_MFD'],
                        dependent_task_list=[flow_accum_task])

                if ('calculate_downstream_distance' in args and
                        bool(args['calculate_downstream_distance'])):
                    distance_path = os.path.join(
                        args['workspace_dir'],
                        _DOWNSTREAM_DISTANCE_FILE_PATTERN % file_suffix)
                    graph.add_task(
                        routing_funcs['distance_to_channel'],
                        args=((flow_dir_path, 1),
                              (stream_mask_path, 1),
                              distance_path),
                        target_path_list=[distance_path],
                        task_name='downstream_distance_%s' % algorithm,
                        dependent_task_list=[stream_threshold_task])
    graph.join()


@validation.invest_validator
def validate(args, limit_to=None):
    """Validate args to ensure they conform to ``execute``'s contract.

    Args:
        args (dict): dictionary of key(str)/value pairs where keys and
            values are specified in ``execute`` docstring.
        limit_to (str): (optional) if not None indicates that validation
            should only occur on the args[limit_to] value. The intent that
            individual key validation could be significantly less expensive
            than validating the entire ``args`` dictionary.

    Returns:
        list of ([invalid key_a, invalid key_b, ...], 'warning/error message')
            tuples. Where an entry indicates that the invalid keys caused
            the error message in the second part of the tuple. This should
            be an empty list if validation succeeds.
    """
    validation_warnings = validation.validate(args, ARGS_SPEC['args'])

    invalid_keys = validation.get_invalid_keys(validation_warnings)
    sufficient_keys = validation.get_sufficient_keys(args)

    if ('dem_band_index' not in invalid_keys and
            'dem_band_index' in sufficient_keys and
            'dem_path' not in invalid_keys and
            'dem_path' in sufficient_keys):
        raster_info = pygeoprocessing.get_raster_info(args['dem_path'])
        if int(args['dem_band_index']) > raster_info['n_bands']:
            validation_warnings.append((
                ['dem_band_index'],
                INVALID_BAND_INDEX_MSG.format(maximum=raster_info['n_bands'])))

    return validation_warnings



# coding=UTF-8

from natcap.invest.ui import model, inputs
from natcap.invest.recreation import recmodel_client
from natcap.invest import MODEL_METADATA


class Recreation(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['recreation'].model_title,
            target=recmodel_client.execute,
            validator=recmodel_client.validate,
            localdoc=MODEL_METADATA['recreation'].userguide)

        self.internet_warning = inputs.Label(
            text=(
                "Note, this computer must have an Internet connection "
                "in order to run this model."))
        self.aoi_path = inputs.File(
            args_key='aoi_path',
            helptext=(
                "An OGR-supported vector file representing the area "
                "of interest where the model will run the analysis."),
            label='Area of Interest (Vector)',
            validator=self.validator)
        self.add_input(self.aoi_path)
        self.start_year = inputs.Text(
            args_key='start_year',
            helptext='Year to start PUD calculations, date starts on Jan 1st.',
            label='Start Year (inclusive, must be >= 2005)',
            validator=self.validator)
        self.add_input(self.start_year)
        self.end_year = inputs.Text(
            args_key='end_year',
            helptext=(
                'Year to end PUD calculations, date ends and includes '
                'Dec 31st.'),
            label='End Year (inclusive, must be <= 2017)',
            validator=self.validator)
        self.add_input(self.end_year)
        self.regression_container = inputs.Container(
            args_key='compute_regression',
            expandable=True,
            expanded=True,
            label='Compute Regression')
        self.add_input(self.regression_container)
        self.predictor_table_path = inputs.File(
            args_key='predictor_table_path',
            helptext=(
                "A table that maps predictor IDs to files and their "
                "types with required headers of 'id', 'path', and "
                "'type'.  The file paths can be absolute, or relative "
                "to the table."),
            label='Predictor Table',
            validator=self.validator)
        self.regression_container.add_input(self.predictor_table_path)
        self.scenario_predictor_table_path = inputs.File(
            args_key='scenario_predictor_table_path',
            helptext=(
                "A table that maps predictor IDs to files and their "
                "types with required headers of 'id', 'path', and "
                "'type'.  The file paths can be absolute, or relative "
                "to the table."),
            label='Scenario Predictor Table (optional)',
            validator=self.validator)
        self.regression_container.add_input(self.scenario_predictor_table_path)
        self.grid_container = inputs.Container(
            args_key='grid_aoi',
            expandable=True,
            expanded=True,
            label='Grid the AOI')
        self.add_input(self.grid_container)
        self.grid_type = inputs.Dropdown(
            args_key='grid_type',
            label='Grid Type',
            options=['square', 'hexagon'])
        self.grid_container.add_input(self.grid_type)
        self.cell_size = inputs.Text(
            args_key='cell_size',
            helptext=(
                "The size of the grid units measured in the "
                "projection units of the AOI. For example, UTM "
                "projections use meters."),
            label='Cell Size',
            validator=self.validator)
        self.grid_container.add_input(self.cell_size)

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.aoi_path.args_key: self.aoi_path.value(),
            self.start_year.args_key: self.start_year.value(),
            self.end_year.args_key: self.end_year.value(),
            self.regression_container.args_key:
                self.regression_container.value(),
            self.grid_container.args_key: self.grid_container.value(),
        }

        if self.regression_container.value():
            args[self.predictor_table_path.args_key] = (
                self.predictor_table_path.value())
            args[self.scenario_predictor_table_path.args_key] = (
                self.scenario_predictor_table_path.value())

        if self.grid_container.value():
            args[self.grid_type.args_key] = self.grid_type.value()
            args[self.cell_size.args_key] = self.cell_size.value()

        return args



# coding=UTF-8
# cython: language_level=2
"""
Implements the Wang et al (2000) viewshed based on reference planes.

This algorithm was originally described in "Generating viewsheds without using
sightlines", authored by Jianjun Wang, Gary J. Robertson, and Kevin White,
published in Photogrammetric Engineering & Remote Sensing, Vol. 66, No. 1,
January 2000, pp. 87-90.

Calculations for adjusting the required height for curvature of the earth have
been adapted from the ESRI ArcGIS documentation:
http://desktop.arcgis.com/en/arcmap/10.3/tools/spatial-analyst-toolbox/using-viewshed-and-observer-points-for-visibility.htm

Consistent with the routing functionality of pygeoprocessing, neighbor
directions follow the right-hand rule where neighbor indexes are interpreted
as:

    # 321
    # 4X0
    # 567
"""
import time
import os
import logging
import shutil
import tempfile

import numpy
import pygeoprocessing
from osgeo import gdal
from osgeo import osr
import shapely.geometry
from .. import utils
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from cython.operator cimport dereference as deref
from cython.operator cimport preincrement as inc
from libc.time cimport time_t
from libc.time cimport time as ctime
from libcpp.list cimport list as clist
from libcpp.set cimport set as cset
from libcpp.deque cimport deque
from libcpp.pair cimport pair
from libcpp.queue cimport queue
from libc cimport math
cimport numpy
cimport cython


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())
BYTE_GTIFF_CREATION_OPTIONS = (
    'GTIFF', ('TILED=YES', 'BIGTIFF=YES', 'COMPRESS=DEFLATE',
              'BLOCKXSIZE=256', 'BLOCKYSIZE=256', 'SPARSE_OK=TRUE'))
FLOAT_GTIFF_CREATION_OPTIONS = (
    'GTIFF', ('PREDICTOR=3',) + BYTE_GTIFF_CREATION_OPTIONS[1])

# Indexes for neighbors relative to the target pixel.
# Indexes in this array are stored in numpy order (row, col).
cdef int* NEIGHBORS_INDEXES = [
    0, 1,  # 0
    -1, 1,  # 1
    -1, 0,  # 2
    -1, -1,  # 3
    0, -1,  # 4
    1, -1,  # 5
    1, 0,  # 6
    1, 1  # 7
]

# The Wang et al algorithm defines two neighbors for the target viewpoint from
# which we construct the reference plane.  In this implementation, Neighbor 1
# is always the neighbor that is closest (in terms of euclidean distance) to
# the target.  Neighbor 2 is at a diagonal to the target.  The indexes of
# Neighbor 1 and Neighbor 2 relative to the target pixel vary by sector.
# This array maps the sector to the index of Neighbor 1, relative to the
# target.
cdef int* SECTOR_TO_NEIGHBOR_1_INDEX = [
    0, -1,
    1, 0,
    1, 0,
    0, 1,
    0, 1,
    -1, 0,
    -1, 0,
    0, -1
]

cdef int* SECTOR_TO_NEIGHBOR_2_INDEX = [
    1, -1,
    1, -1,
    1, 1,
    1, 1,
    -1, 1,
    -1, 1,
    -1, -1,
    -1, -1
]

# List out which targets (in terms of their neighbor index) to process per
# sector.  We can preemptively define which neighbors are to be visited after a
# given target by indexing into the arrays we've already created.  Each line in
# this array has four indexes.  The first two are for Neighbor1, the second two
# are for Neighbor2.
#
# Order of these is:
#    iy_next_neighbor1, ix_next_neighbor1, iy_next_neighbor1, ix_next_neighbor2
#
# When looking at the indexes, the number multiplied by 2 is the neighbor
# index.  So, sector 0 uses neighbors 0 and 1.  Sector 7 uses neighbors 7 and
# 0.
cdef int* SECTOR_NEXT_TARGET_INDEXES = [
    NEIGHBORS_INDEXES[2*0], NEIGHBORS_INDEXES[2*0+1], NEIGHBORS_INDEXES[2*1], NEIGHBORS_INDEXES[2*1+1],
    NEIGHBORS_INDEXES[2*1], NEIGHBORS_INDEXES[2*1+1], NEIGHBORS_INDEXES[2*2], NEIGHBORS_INDEXES[2*2+1],
    NEIGHBORS_INDEXES[2*2], NEIGHBORS_INDEXES[2*2+1], NEIGHBORS_INDEXES[2*3], NEIGHBORS_INDEXES[2*3+1],
    NEIGHBORS_INDEXES[2*3], NEIGHBORS_INDEXES[2*3+1], NEIGHBORS_INDEXES[2*4], NEIGHBORS_INDEXES[2*4+1],
    NEIGHBORS_INDEXES[2*4], NEIGHBORS_INDEXES[2*4+1], NEIGHBORS_INDEXES[2*5], NEIGHBORS_INDEXES[2*5+1],
    NEIGHBORS_INDEXES[2*5], NEIGHBORS_INDEXES[2*5+1], NEIGHBORS_INDEXES[2*6], NEIGHBORS_INDEXES[2*6+1],
    NEIGHBORS_INDEXES[2*6], NEIGHBORS_INDEXES[2*6+1], NEIGHBORS_INDEXES[2*7], NEIGHBORS_INDEXES[2*7+1],
    NEIGHBORS_INDEXES[2*7], NEIGHBORS_INDEXES[2*7+1], NEIGHBORS_INDEXES[2*0], NEIGHBORS_INDEXES[2*0+1]
]


# The processing queue must be seeded with a pixel to process.  From there, the
# processing loop will discover more neighbors to inspect until it runs out of
# neighbors in the sector.  These seed indexes are relative to the viewpoint
# index.
cdef int* SECTOR_SEEDS = [
    -1, 2,  # seed index for sector 0
    -2, 1,  # seed index for sector 1
    -2, -1,  # seed index for sector 2
    -1, -2,  # seed index for sector 3
    1, -2,  # seed index for sector 4
    2, -1,  # seed index for sector 5
    2, 1,  # seed index for sector 6
    1, 2,  # seed index for sector 7
]
cdef int AUX_NOT_VISITED = -9999
cdef int DIAM_EARTH = 12740000  # meters, from ArcGIS docs
cdef double DIAM_EARTH_INV = 1.0/DIAM_EARTH
cdef double IMPROBABLE_NODATA = -123457.12345

# This type represents an x,y coordinate pair.
ctypedef pair[long, long] CoordinatePair

# Defining a couple of operators for longs so I can minimize code to be
# executed based on `cython -a`.
cdef inline long labs(long a):
    if a > 0:
        return a
    return a*-1

cdef inline long lmax(long a, long b):
    if a > b:
        return a
    return b

cdef inline long lmin(long a, long b):
    if a < b:
        return a
    return b


# this is a least recently used cache written in C++ in an external file,
# exposing here so _ManagedRaster can use it.
# Copied from src/pygeoprocessing/routing/routing.pyx,
# revision 891288683889237cfd3a3d0a1f09483c23489fca.
cdef extern from "LRUCache.h":
    cdef cppclass LRUCache[KEY_T, VAL_T]:
        LRUCache(int)
        void put(KEY_T&, VAL_T&, clist[pair[KEY_T,VAL_T]]&)
        clist[pair[KEY_T,VAL_T]].iterator begin()
        clist[pair[KEY_T,VAL_T]].iterator end()
        bint exist(KEY_T &)
        VAL_T get(KEY_T &)


# exposing stl::priority_queue so we can have all 3 template arguments so
# we can pass a different Compare functor
cdef extern from "<queue>" namespace "std":
    cdef cppclass priority_queue[T, Container, Compare]:
        priority_queue() except +
        priority_queue(priority_queue&) except +
        priority_queue(Container&)
        bint empty()
        void pop()
        void push(T&)
        size_t size()
        T& top()


# functor for priority queue of pixels
cdef cppclass BlockwiseCloserTarget:
    bint get "operator()"(TargetPixel& lhs, TargetPixel& rhs):
        # Go with the lower ring ID if possible.
        if lhs.ring_id < rhs.ring_id:
            return 0

        # If we're in the same ring, go with the lowest sector
        # If we're in the same sector, go with the shortest distance.
        if lhs.ring_id == rhs.ring_id:
            if lhs.sector < rhs.sector:
                return 0
            if lhs.sector == rhs.sector:
                if lhs.distance_to_viewpoint < rhs.distance_to_viewpoint:
                    return 0
        return 1


# A priority queue of TargetPixels should should first minimize the block distance.
# If the block_distance is the same, the queue should minimize the target distance.
cdef struct TargetPixel:
    long ix
    long iy
    int ring_id
    int sector
    double distance_to_viewpoint # distance between viewpoint and target.


ctypedef priority_queue[
    TargetPixel, deque[TargetPixel], BlockwiseCloserTarget] TargetPixelPriorityQueue

# A function for wrapping up a call to get the length of the hypotenuse between two pixels.
cdef inline double pixel_dist(long ix_source, long ix_target, long iy_source, long iy_target):
        return math.hypot(
            lmax(ix_source, ix_target)-lmin(ix_source, ix_target),
            lmax(iy_source, iy_target)-lmin(iy_source, iy_target))


# Number of raster blocks to hold in memory at once per Managed Raster
cdef int MANAGED_RASTER_N_BLOCKS = 2**6

# The nodata value for visibility rasters
cdef int VISIBILITY_NODATA = 255

# this ctype is used to store the block ID and the block buffer as one object
# inside Managed Raster
ctypedef pair[int, double*] BlockBufferPair

# a class to allow fast random per-pixel access to a raster for both setting
# and reading pixels.  Copied from src/pygeoprocessing/routing/routing.pyx,
# revision 891288683889237cfd3a3d0a1f09483c23489fca.
cdef class _ManagedRaster:
    cdef LRUCache[int, double*]* lru_cache
    cdef cset[int] dirty_blocks
    cdef int block_xsize
    cdef int block_ysize
    cdef int block_xmod
    cdef int block_ymod
    cdef int block_xbits
    cdef int block_ybits
    cdef long raster_x_size
    cdef long raster_y_size
    cdef int block_nx
    cdef int block_ny
    cdef int write_mode
    cdef bytes raster_path
    cdef int band_id
    cdef int closed

    def __cinit__(self, raster_path, band_id, write_mode):
        """Create new instance of Managed Raster.

        Args:
            raster_path (char*): path to raster that has block sizes that are
                powers of 2. If not, an exception is raised.
            band_id (int): which band in `raster_path` to index. Uses GDAL
                notation that starts at 1.
            write_mode (boolean): if true, this raster is writable and dirty
                memory blocks will be written back to the raster as blocks
                are swapped out of the cache or when the object deconstructs.

        Returns:
            None.
        """
        raster_info = pygeoprocessing.get_raster_info(raster_path)
        self.raster_x_size, self.raster_y_size = raster_info['raster_size']
        self.block_xsize, self.block_ysize = raster_info['block_size']
        self.block_xmod = self.block_xsize-1
        self.block_ymod = self.block_ysize-1

        if not (1 <= band_id <= raster_info['n_bands']):
            err_msg = (
                "Error: band ID (%s) is not a valid band number. "
                "This exception is happening in Cython, so it will cause a "
                "hard seg-fault, but it's otherwise meant to be a "
                "ValueError." % (band_id))
            print(err_msg)
            raise ValueError(err_msg)
        self.band_id = band_id

        if (self.block_xsize & (self.block_xsize - 1) != 0) or (
                self.block_ysize & (self.block_ysize - 1) != 0):
            # If inputs are not a power of two, this will at least print
            # an error message. Unfortunately with Cython, the exception will
            # present itself as a hard seg-fault, but I'm leaving the
            # ValueError in here at least for readability.
            err_msg = (
                "Error: Block size is not a power of two: "
                "block_xsize: %d, %d, %s. This exception is happening"
                "in Cython, so it will cause a hard seg-fault, but it's"
                "otherwise meant to be a ValueError." % (
                    self.block_xsize, self.block_ysize, raster_path))
            print(err_msg)
            raise ValueError(err_msg)

        self.block_xbits = numpy.log2(self.block_xsize)
        self.block_ybits = numpy.log2(self.block_ysize)
        self.block_nx = (
            self.raster_x_size + (self.block_xsize) - 1) / self.block_xsize
        self.block_ny = (
            self.raster_y_size + (self.block_ysize) - 1) / self.block_ysize

        self.lru_cache = new LRUCache[int, double*](MANAGED_RASTER_N_BLOCKS)
        self.raster_path = <bytes> raster_path
        self.write_mode = write_mode
        self.closed = 0

    def __dealloc__(self):
        """Deallocate _ManagedRaster.

        This operation manually frees memory from the LRUCache and writes any
        dirty memory blocks back to the raster if `self.write_mode` is True.
        """
        self.close()

    def close(self):
        """Close the _ManagedRaster and free up resources.

            This call writes any dirty blocks to disk, frees up the memory
            allocated as part of the cache, and frees all GDAL references.

            Any subsequent calls to any other functions in _ManagedRaster will
            have undefined behavior.
        """
        if self.closed:
            return
        self.closed = 1
        cdef int xi_copy, yi_copy
        cdef numpy.ndarray[double, ndim=2] block_array = numpy.empty(
            (self.block_ysize, self.block_xsize))
        cdef double *double_buffer
        cdef int block_xi
        cdef int block_yi
        # initially the win size is the same as the block size unless
        # we're at the edge of a raster
        cdef int win_xsize
        cdef int win_ysize

        # we need the offsets to subtract from global indexes for cached array
        cdef int xoff
        cdef int yoff

        cdef clist[BlockBufferPair].iterator it = self.lru_cache.begin()
        cdef clist[BlockBufferPair].iterator end = self.lru_cache.end()
        if not self.write_mode:
            while it != end:
                # write the changed value back if desired
                PyMem_Free(deref(it).second)
                inc(it)
            return

        raster = gdal.OpenEx(
            self.raster_path, gdal.GA_Update | gdal.OF_RASTER)
        raster_band = raster.GetRasterBand(self.band_id)

        # if we get here, we're in write_mode
        cdef cset[int].iterator dirty_itr
        while it != end:
            double_buffer = deref(it).second
            block_index = deref(it).first

            # write to disk if block is dirty
            dirty_itr = self.dirty_blocks.find(block_index)
            if dirty_itr != self.dirty_blocks.end():
                self.dirty_blocks.erase(dirty_itr)
                block_xi = block_index % self.block_nx
                block_yi = block_index / self.block_nx

                # we need the offsets to subtract from global indexes for
                # cached array
                xoff = block_xi << self.block_xbits
                yoff = block_yi << self.block_ybits

                win_xsize = self.block_xsize
                win_ysize = self.block_ysize

                # clip window sizes if necessary
                if xoff+win_xsize > self.raster_x_size:
                    win_xsize = win_xsize - (
                        xoff+win_xsize - self.raster_x_size)
                if yoff+win_ysize > self.raster_y_size:
                    win_ysize = win_ysize - (
                        yoff+win_ysize - self.raster_y_size)

                for xi_copy in xrange(win_xsize):
                    for yi_copy in xrange(win_ysize):
                        block_array[yi_copy, xi_copy] = (
                            double_buffer[
                                (yi_copy << self.block_xbits) + xi_copy])
                raster_band.WriteArray(
                    block_array[0:win_ysize, 0:win_xsize],
                    xoff=xoff, yoff=yoff)
            PyMem_Free(double_buffer)
            inc(it)
        raster_band.FlushCache()
        raster_band = None
        raster = None

    cdef inline void set(self, long xi, long yi, double value):
        """Set the pixel at `xi,yi` to `value`."""
        cdef int block_xi = xi >> self.block_xbits
        cdef int block_yi = yi >> self.block_ybits
        # this is the flat index for the block
        cdef int block_index = block_yi * self.block_nx + block_xi
        if not self.lru_cache.exist(block_index):
            self._load_block(block_index)
        self.lru_cache.get(
            block_index)[
                ((yi & (self.block_ymod))<<self.block_xbits) +
                (xi & (self.block_xmod))] = value
        if self.write_mode:
            dirty_itr = self.dirty_blocks.find(block_index)
            if dirty_itr == self.dirty_blocks.end():
                self.dirty_blocks.insert(block_index)

    cdef inline double get(self, long xi, long yi):
        """Return the value of the pixel at `xi,yi`."""
        cdef int block_xi = xi >> self.block_xbits
        cdef int block_yi = yi >> self.block_ybits
        # this is the flat index for the block
        cdef int block_index = block_yi * self.block_nx + block_xi
        if not self.lru_cache.exist(block_index):
            self._load_block(block_index)
        return self.lru_cache.get(
            block_index)[
                ((yi & (self.block_ymod))<<self.block_xbits) +
                (xi & (self.block_xmod))]

    cdef void _load_block(self, int block_index) except *:
        cdef int block_xi = block_index % self.block_nx
        cdef int block_yi = block_index / self.block_nx

        # we need the offsets to subtract from global indexes for cached array
        cdef int xoff = block_xi << self.block_xbits
        cdef int yoff = block_yi << self.block_ybits

        cdef int xi_copy, yi_copy
        cdef numpy.ndarray[double, ndim=2] block_array
        cdef double *double_buffer
        cdef clist[BlockBufferPair] removed_value_list

        # determine the block aligned xoffset for read as array

        # initially the win size is the same as the block size unless
        # we're at the edge of a raster
        cdef int win_xsize = self.block_xsize
        cdef int win_ysize = self.block_ysize

        # load a new block
        if xoff+win_xsize > self.raster_x_size:
            win_xsize = win_xsize - (xoff+win_xsize - self.raster_x_size)
        if yoff+win_ysize > self.raster_y_size:
            win_ysize = win_ysize - (yoff+win_ysize - self.raster_y_size)

        raster = gdal.OpenEx(self.raster_path, gdal.OF_RASTER)
        raster_band = raster.GetRasterBand(self.band_id)
        block_array = raster_band.ReadAsArray(
            xoff=xoff, yoff=yoff, win_xsize=win_xsize,
            win_ysize=win_ysize).astype(
            numpy.float64)
        raster_band = None
        raster = None
        double_buffer = <double*>PyMem_Malloc(
            (sizeof(double) << self.block_xbits) * win_ysize)
        for xi_copy in xrange(win_xsize):
            for yi_copy in xrange(win_ysize):
                double_buffer[(yi_copy<<self.block_xbits)+xi_copy] = (
                    block_array[yi_copy, xi_copy])
        self.lru_cache.put(
            <int>block_index, <double*>double_buffer, removed_value_list)

        if self.write_mode:
            raster = gdal.OpenEx(
                self.raster_path, gdal.GA_Update | gdal.OF_RASTER)
            raster_band = raster.GetRasterBand(self.band_id)

        block_array = numpy.empty(
            (self.block_ysize, self.block_xsize), dtype=numpy.double)
        while not removed_value_list.empty():
            # write the changed value back if desired
            double_buffer = removed_value_list.front().second

            if self.write_mode:
                block_index = removed_value_list.front().first

                # write back the block if it's dirty
                dirty_itr = self.dirty_blocks.find(block_index)
                if dirty_itr != self.dirty_blocks.end():
                    self.dirty_blocks.erase(dirty_itr)

                    block_xi = block_index % self.block_nx
                    block_yi = block_index / self.block_nx

                    xoff = block_xi << self.block_xbits
                    yoff = block_yi << self.block_ybits

                    win_xsize = self.block_xsize
                    win_ysize = self.block_ysize

                    if xoff+win_xsize > self.raster_x_size:
                        win_xsize = win_xsize - (
                            xoff+win_xsize - self.raster_x_size)
                    if yoff+win_ysize > self.raster_y_size:
                        win_ysize = win_ysize - (
                            yoff+win_ysize - self.raster_y_size)

                    for xi_copy in xrange(win_xsize):
                        for yi_copy in xrange(win_ysize):
                            block_array[yi_copy, xi_copy] = double_buffer[
                                (yi_copy << self.block_xbits) + xi_copy]
                    raster_band.WriteArray(
                        block_array[0:win_ysize, 0:win_xsize],
                        xoff=xoff, yoff=yoff)
            PyMem_Free(double_buffer)
            removed_value_list.pop_front()

        if self.write_mode:
            raster_band = None
            raster = None


@cython.binding(True)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def viewshed(dem_raster_path_band,
             viewpoint,
             visibility_filepath,
             viewpoint_height=0.0,
             curved_earth=True,
             refraction_coeff=0.13,
             max_distance=None,
             aux_filepath=None):
    """Compute the Wang et al. reference-plane based viewshed.

    Args:
        dem_raster_path_band (tuple): A tuple of (path, band_index) where
            ``path`` is a path to a GDAL-compatible raster on disk and
            ``band_index`` is the 1-based band index.  This DEM must be tiled
            with block sizes as a power of 2.  If the viewshed is being
            adjusted for curvature of the earth and/or refraction, the
            elevation units of the DEM must be in meters.  The DEM need not be
            projected in meters.
        viewpoint (tuple):  A tuple of 2 numbers in the order
            ``(east offset, north offset)``.  These units must be of the same
            units as the coordinate system of the DEM.  This index represents
            the viewpoint location.  The closest pixel to this viewpoint index
            will be used as the viewpoint.
        visibility_filepath (string): A filepath on disk to where the
            visibility raster will be written. If a raster exists in this
            location, it will be overwritten.
        viewpoint_height=0.0 (float):  The height (in the units of the DEM
            height) of the observer at the viewpoint.
        curved_earth=True (bool): Whether to adjust viewshed calculations for
            the curvature of the earth.  If False, the earth will be treated as
            though it is flat.
        refraction_coeff=0.13 (float):  The coefficient of atmospheric
            refraction that may be adjusted to accommodate varying atmospheric
            conditions.  Default is ``0.13``.  Set to ``0`` to ignore
            refraction calculations.
        max_distance=None (float):  If provided, visibility will not be
            calculated for DEM pixels that are more than this distance (in meters)
            from the viewpoint.
        aux_filepath=None (string): A path to a location on disk
            where the raster containing the auxiliary matrix will be written.
            The auxiliary matrix defines the height that a DEM must exceed in
            order to be visible from the viewpoint.  This matrix is very useful
            for debugging.  If a raster already exists at this location, it
            will be overwritten.  If this path is not provided by the user, the
            viewshed will create this file as a temporary file wherever the
            system keeps its temp files and remove it when the viewshed
            finishes.  See python's ``tempfile`` documentation for where this
            might be on your system.

    Raises:
        ValueError: When either the viewpoint does not overlap with the DEM or
            the DEM is not tiled appropriately.

        LookupError: When the ``viewpoint`` coordinate pair is over nodata.

        AssertionError: When pixel dimensions are not square.

    Returns:
        ``None``
    """
    start_time = time.time()

    # Check the bounding box to make sure that the viewpoint overlaps the DEM.
    dem_raster_info = pygeoprocessing.get_raster_info(dem_raster_path_band[0])
    dem_gt = dem_raster_info['geotransform']
    bbox_minx, bbox_miny, bbox_maxx, bbox_maxy = dem_raster_info['bounding_box']
    if (not bbox_minx <= viewpoint[0] <= bbox_maxx or
            not bbox_miny <= viewpoint[1] <= bbox_maxy):
        raise ValueError(('Viewpoint (%s, %s) does not overlap with DEM with '
                          'bounding box %s') % (viewpoint[0], viewpoint[1],
                                                dem_raster_info['bounding_box']))
    cdef long iy_viewpoint = int((viewpoint[1] - dem_gt[3]) / dem_gt[5])
    cdef long ix_viewpoint = int((viewpoint[0] - dem_gt[0]) / dem_gt[1])

    # Get the elevation of the pixel under the viewpoint and check to see if
    # it's nodata.
    raster = gdal.OpenEx(dem_raster_path_band[0])
    band = raster.GetRasterBand(dem_raster_path_band[1])
    viewpoint_elevation = band.ReadAsArray(ix_viewpoint, iy_viewpoint, 1, 1)
    band = None
    raster = None

    band_to_array_index = dem_raster_path_band[1] - 1
    nodata_value = dem_raster_info['nodata'][band_to_array_index]
    if viewpoint_elevation == nodata_value:
        raise LookupError('Viewpoint is over nodata')

    # Need to handle the case where the nodata value is not defined.
    if nodata_value is None:
        nodata_value = IMPROBABLE_NODATA
    cdef double nodata = nodata_value

    # Verify that pixels are very close to square.  The Wang et al algorithm
    # doesn't require that this be the case, but the math is simplified if it
    # is.
    pixel_xsize, pixel_ysize = dem_raster_info['pixel_size']
    if not (abs(abs(pixel_xsize) - abs(pixel_ysize)) < 0.5e-7):
        raise AssertionError(
            'Pixel dimensions must match:\n X size:%s\n Y size:%s' %
                             (pixel_xsize, pixel_ysize))

    # Verify that the block sizes are powers of 2 and are square.
    # This is needed for the _ManagedRaster classes.  If this is not asserted
    # here, the _ManagedRaster classes will crash with a segfault.
    block_xsize, block_ysize = dem_raster_info['block_size']
    if (block_xsize & (block_xsize - 1) != 0 or (
            block_ysize & (block_ysize - 1) != 0)) or (
                block_xsize != block_ysize):
        raise ValueError(
            'DEM must be tiled and tiles must be of equal sizes that are a '
            'power of 2.  Current block size is (%s, %s)' %
            (block_xsize, block_ysize))

    # Create the auxiliary raster for storing the calculated minimum height
    # for visibility at a given point.
    temp_dir = None
    if aux_filepath is None:
        temp_dir = tempfile.mkdtemp(
            prefix='viewshed_%s' % time.strftime(
                '%Y-%m-%d_%H_%M_%S', time.gmtime()))
        aux_filepath = os.path.join(temp_dir, 'auxiliary.tif')

    LOGGER.info("Creating auxiliary raster %s", aux_filepath)
    pygeoprocessing.new_raster_from_base(
        dem_raster_path_band[0], aux_filepath, gdal.GDT_Float64, [AUX_NOT_VISITED],
        fill_value_list=[AUX_NOT_VISITED],
        raster_driver_creation_tuple=FLOAT_GTIFF_CREATION_OPTIONS)

    # Create the visibility raster for indicating whether a pixel is visible
    # based on the calculated minimum height.
    LOGGER.info('Creating visibility raster %s', visibility_filepath)
    pygeoprocessing.new_raster_from_base(
        dem_raster_path_band[0], visibility_filepath, gdal.GDT_Byte,
        [VISIBILITY_NODATA], fill_value_list=[VISIBILITY_NODATA],
        raster_driver_creation_tuple=BYTE_GTIFF_CREATION_OPTIONS)

    # LRU-cached rasters for easier access to individual pixels.
    cdef _ManagedRaster dem_managed_raster = (
            _ManagedRaster(dem_raster_path_band[0], dem_raster_path_band[1], 0))
    cdef _ManagedRaster aux_managed_raster = (
            _ManagedRaster(aux_filepath, 1, 1))
    cdef _ManagedRaster visibility_managed_raster = (
            _ManagedRaster(visibility_filepath, 1, 1))

    # get the pixel size in terms of meters.
    dem_srs = osr.SpatialReference()
    dem_srs.ImportFromWkt(dem_raster_info['projection_wkt'])
    linear_units = dem_srs.GetLinearUnits()
    cdef double pixel_size = utils.mean_pixel_size_and_area(
        dem_raster_info['pixel_size'])[0]*linear_units
    cdef long raster_x_size = dem_raster_info['raster_size'][0]
    cdef long raster_y_size = dem_raster_info['raster_size'][1]
    cdef double max_visible_radius
    cdef long pixels_in_raster
    if max_distance is not None:
        # This is an estimate of the number of pixels in the raster when
        # viewshed is given a max_distance radius.  The estimate is based on
        # the area of the DEM bounding box that intersects with the circle
        # indicating the max_distance radius around the viewpoint.
        max_visible_radius = max_distance

        # prefer to overestimate.
        point = shapely.geometry.Point(viewpoint).buffer(max_distance+pixel_size)
        bbox_rectangle = shapely.geometry.box(*dem_raster_info['bounding_box'])
        pixels_in_raster = <long> int(
            point.intersection(bbox_rectangle).area/(pixel_size**2))
        LOGGER.debug('Approx. %s pixels in raster within max distance %s',
                     pixels_in_raster, max_distance)
    else:
        # max visible distance is the whole raster, which should be the hypotenuse
        # between two adjoining edges of the bounding box.
        max_visible_radius = math.hypot(raster_x_size, raster_y_size)*pixel_size
        pixels_in_raster = raster_x_size * raster_y_size

    cdef long m, n, xi, yi
    cdef int correct_for_curvature = curved_earth
    cdef int correct_for_refraction = math.fabs(math.ceil(refraction_coeff) - 1.0) < 0.5e-7
    cdef double target_height_adjustment = 0  # initializing for compiler
    cdef double adjustment = 0.0
    cdef float refract_coeff = refraction_coeff  # from the user
    cdef int block_x_size = dem_raster_info['block_size'][0]
    cdef int block_y_size = dem_raster_info['block_size'][1]
    cdef int block_bits = numpy.log2(block_x_size)  # for bit-shifting
    cdef long ix_viewpoint_block = ix_viewpoint >> block_x_size
    cdef long iy_viewpoint_block = iy_viewpoint >> block_y_size
    cdef long pixels_touched = 0

    # Following the Wang et al. terminology, it's helpful to think of blocks in
    # terms of rings around the block containing the viewpoint.  Ring 0 is the
    # 'ring' containing the viewpoint.  A block in ring 1 is one of the 8
    # blocks immediately adjacent to the viewpoint's block.
    cdef int ring_id

    LOGGER.info("Starting viewshed for viewpoint %s on DEM %s",
                viewpoint, dem_raster_path_band[0])

    # As defined by Wang et al, the viewpoint and the immediate neighbors are
    # all assumed to be visible.
    for yi in xrange(iy_viewpoint-1, iy_viewpoint+2):
        # Bounds check on the row we're initializing
        if not 0 <= yi < raster_y_size:
            continue

        for xi in xrange(ix_viewpoint-1, ix_viewpoint+2):
            # Bounds check on the column we're initializing.
            if not 0 <= xi < raster_x_size:
                continue

            aux_managed_raster.set(
                xi, yi, dem_managed_raster.get(xi, yi))
            visibility_managed_raster.set(xi, yi, 1)
            pixels_touched += 1

    # Save the viewpoint index for later.  These are the variable names used in
    # the reference-plane equation in Wang et al.
    # i is the row, j is the column.
    cdef long i = iy_viewpoint
    cdef long j = ix_viewpoint

    # If the user defined a viewpoint, we add it to the actual viewpoint height
    # in the DEM matrix.
    cdef double r_v = dem_managed_raster.get(ix_viewpoint, iy_viewpoint) + viewpoint_height

    # Defining cardinal and intercardinal directions is significantly simpler
    # than defining other pixels because the reference plane is constructed
    # from only 1 previous pixel (as opposed to 2 for the other pixels).
    # Rather than try to shoehorn these calculations into a single processing
    # loop, we can take care of this special case ahead of time.
    cdef long ix_target, iy_target
    cdef long ix_prev_target, iy_prev_target
    cdef long ix_cardinal_target, iy_cardinal_target
    cdef double target_dem_height, adjusted_dem_height
    cdef double target_distance
    cdef double slope_distance
    cdef double z = 0  # initializing for compiler
    cdef int multiplier
    cdef double sqrt2 = math.sqrt(2)
    cdef int i_n
    cdef time_t last_log_time = ctime(NULL)
    for i_n in xrange(8):
        ix_cardinal_target = NEIGHBORS_INDEXES[2*i_n]
        iy_cardinal_target = NEIGHBORS_INDEXES[2*i_n+1]
        multiplier = 2
        while True:
            iy_target = iy_viewpoint+iy_cardinal_target*multiplier
            if not 0 <= iy_target < raster_y_size:
                break

            ix_target = ix_viewpoint+ix_cardinal_target*multiplier
            if not 0 <= ix_target < raster_x_size:
                break

            ix_prev_target = ix_viewpoint+ix_cardinal_target*(multiplier-1)
            iy_prev_target = iy_viewpoint+iy_cardinal_target*(multiplier-1)
            previous_height = aux_managed_raster.get(ix_prev_target,
                                                     iy_prev_target)

            if lmax(ix_cardinal_target, iy_cardinal_target) == 0:
                slope_distance = labs(
                    lmin(ix_cardinal_target, iy_cardinal_target)*(multiplier-1))
                target_distance = labs(
                    lmin(ix_cardinal_target, iy_cardinal_target)*(multiplier))
            else:
                slope_distance = labs(
                    lmax(ix_cardinal_target, iy_cardinal_target)*(multiplier-1))
                target_distance = labs(
                    lmax(ix_cardinal_target, iy_cardinal_target)*(multiplier))

            # If we're on a diagonal, multiply by sqrt(2) to adjust the pixel
            # distance for the diagonal.
            if ix_cardinal_target != 0 and iy_cardinal_target != 0:
                slope_distance *= sqrt2
                target_distance *= sqrt2

            # adjust for pixel_size
            target_distance *= pixel_size
            slope_distance *= pixel_size

            if target_distance > max_visible_radius:
                break

            z = (((previous_height-r_v)/slope_distance) *
                 target_distance + r_v)

            # add on refractivity/curvature-of-earth calculations.
            adjustment = 0.0  # increase in required height due to curvature
            if correct_for_curvature or correct_for_refraction:
                # target_height_adjustment is the apparent height reduction of
                # the target due to the curvature of the earth.
                target_height_adjustment = (target_distance**2)*DIAM_EARTH_INV
                if correct_for_curvature:
                    adjustment += target_height_adjustment
                if correct_for_refraction:
                    adjustment -= refract_coeff*target_height_adjustment

            target_dem_height = dem_managed_raster.get(ix_target, iy_target)
            adjusted_dem_height = target_dem_height - adjustment
            if (adjusted_dem_height >= z and
                    target_distance < max_visible_radius and
                    target_dem_height != nodata):
                visibility_managed_raster.set(ix_target, iy_target, 1)
                aux_managed_raster.set(ix_target, iy_target, adjusted_dem_height)
            else:
                visibility_managed_raster.set(ix_target, iy_target, 0)
                aux_managed_raster.set(ix_target, iy_target, z)

            multiplier += 1
            pixels_touched += 1

    # Process subsequent land points.
    # The queue keeps track of the order, the set tracks what's in the queue.
    # The set allows us to do constant-time checking of what's in the queue so
    # we don't add more work than we have to.
    cdef TargetPixelPriorityQueue process_queue
    cdef cset[CoordinatePair] process_queue_set

    cdef int next_target_idx
    cdef long ix_seed, iy_seed
    cdef int sector
    cdef long ix_next_target, iy_next_target
    cdef CoordinatePair next_target_index
    cdef TargetPixel target_pixel
    cdef int pixels_touched_at_last_log = 0
    cdef int pixels_processed_since_last_log

    # Seed the sector with a pixel in the same direction as the sector.
    # We can only seed the pixel if it's within the raster.
    for sector in xrange(0, 8):
        iy_seed = iy_viewpoint + SECTOR_SEEDS[2*sector]
        if not 0 <= iy_seed < raster_y_size:
            continue

        ix_seed = ix_viewpoint + SECTOR_SEEDS[2*sector+1]
        if not 0 <= ix_seed < raster_x_size:
            continue

        target_distance = pixel_dist(ix_viewpoint, ix_seed,
                                     iy_viewpoint, iy_seed)*pixel_size
        ring_id = lmax(labs(ix_viewpoint_block - ix_seed>>block_bits),
                       labs(iy_viewpoint_block - iy_seed>>block_bits))

        process_queue.push(TargetPixel(
            ix_seed, iy_seed, ring_id, sector, target_distance))
        process_queue_set.insert(CoordinatePair(ix_seed, iy_seed))

    while not process_queue.empty():
        if ctime(NULL) - last_log_time > 5.0:
            pixels_processed_since_last_log = pixels_touched - pixels_touched_at_last_log
            time_since_last_log = ctime(NULL) - last_log_time
            last_log_time = ctime(NULL)
            LOGGER.info(
                ('Viewshed approx. %.2f%% complete. Remaining pixels: %i '
                 '(%6.2f pixels/sec)'),
                100.0*pixels_touched/<double>pixels_in_raster,
                pixels_in_raster-pixels_touched,
                pixels_processed_since_last_log/<double>time_since_last_log
            )
            pixels_touched_at_last_log = pixels_touched

        target_pixel = process_queue.top()
        process_queue_set.erase(
            CoordinatePair(target_pixel.ix, target_pixel.iy))
        process_queue.pop()

        # We have a target, determine visibility
        m = target_pixel.iy  # y index (row)
        n = target_pixel.ix   # x index (col)
        r_n1 = aux_managed_raster.get(
            n + SECTOR_TO_NEIGHBOR_1_INDEX[2*target_pixel.sector+1],
            m + SECTOR_TO_NEIGHBOR_1_INDEX[2*target_pixel.sector])
        r_n2 = aux_managed_raster.get(
            n + SECTOR_TO_NEIGHBOR_2_INDEX[2*target_pixel.sector+1],
            m + SECTOR_TO_NEIGHBOR_2_INDEX[2*target_pixel.sector])

        # These equations are taken directly from the Wang et al. paper.
        # Sector 3 is the sector that is explicitly referenced in the paper.
        # The others are adjusted based on the neighbors selected.
        # I could abstract this in such a way that the math is only written out
        # on one line, but that ends up being much longer than just writing out
        # the math here for each sector.
        #
        # Variable names in the math has been maintained where possible.
        # r_n1 is the equivalent of r_m,n+1 in the paper.
        # r_n2 is the equivalent of r+m+1,n+1 in the paper.
        #
        # The only modification I have made to the math is working in the
        # viewpoint height, r_v, which allows us to have a slope that is
        # positive or negative, and is relative to the viewpoint height.  This
        # modification is not in the paper, and is not in any citable resource
        # I have found.
        if target_pixel.sector == 0:
            z = -(m-i)*(r_n1-r_n2)+(j-n)*((m-i)*(r_n1-r_n2)-r_v+r_n1)/(j+1-n)+r_v
        elif target_pixel.sector == 1:
            z = -(j-n)*(r_n1-r_n2)+(m-i)*((j-n)*(r_n1-r_n2)-r_v+r_n1)/(m+1-i)+r_v
        elif target_pixel.sector == 2:
            z = -(n-j)*(r_n1-r_n2)+(m-i)*((n-j)*(r_n1-r_n2)-r_v+r_n1)/(m+1-i)+r_v
        elif target_pixel.sector == 3:
            z = -(m-i)*(r_n1-r_n2)+(n-j)*((m-i)*(r_n1-r_n2)-r_v+r_n1)/(n+1-j)+r_v
        elif target_pixel.sector == 4:
            z = -(i-m)*(r_n1-r_n2)+(n-j)*((i-m)*(r_n1-r_n2)-r_v+r_n1)/(n+1-j)+r_v
        elif target_pixel.sector == 5:
            z = -(n-j)*(r_n1-r_n2)+(i-m)*((n-j)*(r_n1-r_n2)-r_v+r_n1)/(i+1-m)+r_v
        elif target_pixel.sector == 6:
            z = -(j-n)*(r_n1-r_n2)+(i-m)*((j-n)*(r_n1-r_n2)-r_v+r_n1)/(i+1-m)+r_v
        elif target_pixel.sector == 7:
            z = -(i-m)*(r_n1-r_n2)+(j-n)*((i-m)*(r_n1-r_n2)-r_v+r_n1)/(j+1-n)+r_v

        # Refraction and curvature calculations are not in the Wang et al
        # paper.  I adapted these from the ESRI documentation of their
        # viewshed, which can be found at
        # http://desktop.arcgis.com/en/arcmap/10.3/tools/spatial-analyst-toolbox/using-viewshed-and-observer-points-for-visibility.htm
        adjustment = 0.0  # increase in required height due to curvature
        if correct_for_curvature or correct_for_refraction:
            # target_height_adjustment is the apparent height reduction of
            # the target due to the curvature of the earth.
            target_height_adjustment = (target_pixel.distance_to_viewpoint**2)*DIAM_EARTH_INV
            if correct_for_curvature:
                adjustment += target_height_adjustment
            if correct_for_refraction:
                adjustment -= refract_coeff*target_height_adjustment

        # Given the reference plane and any adjustments to the minimum required
        # height for visibility, the DEM pixel is only visible if it is greater
        # than or equal to the minimum-visible height AND is closer than the
        # maximum visible radius.
        target_dem_height = dem_managed_raster.get(n, m)
        adjusted_dem_height = dem_managed_raster.get(n, m) - adjustment
        if (adjusted_dem_height >= z and
                target_pixel.distance_to_viewpoint < max_visible_radius and
                target_dem_height != nodata):
            visibility_managed_raster.set(n, m, 1)
            aux_managed_raster.set(n, m, adjusted_dem_height)
        else:
            # If it's close enough to nodata to be interpreted as nodata,
            # consider it to be nodata.  Nodata implies that visibility is
            # undefined ... which it is, since there's no defined DEM value for
            # this pixel.
            if math.fabs(target_dem_height - nodata) <= 1.0e-7:
                visibility_managed_raster.set(n, m, VISIBILITY_NODATA)
            else:
                # If we're not over nodata, then the pixel isn't visible.
                visibility_managed_raster.set(n, m, 0)
            aux_managed_raster.set(n, m, z)
        pixels_touched += 1

        # Having determined the visibility for the target_pixel, we now need to
        # enqueue those pixels that depend on this pixel.  This is determined
        # by the sector that target_pixel is in.  Every target has two possible
        # neighbors that may be enqueued, but there are a variety of reasons
        # why a neighbor might not be a valid target.
        for next_target_idx in xrange(target_pixel.sector*4,
                                      target_pixel.sector*4+4, 2):
            iy_next_target = m + SECTOR_NEXT_TARGET_INDEXES[next_target_idx]
            ix_next_target = n + SECTOR_NEXT_TARGET_INDEXES[next_target_idx+1]

            # Skip the target if it's off the bounds of the raster.
            if not 0 <= iy_next_target < raster_y_size:
                continue
            if not 0 <= ix_next_target < raster_x_size:
                continue

            # Skip the target if it's already in the queue.
            next_target_index = CoordinatePair(ix_next_target, iy_next_target)
            if (process_queue_set.find(next_target_index)
                    != process_queue_set.end()):
                continue

            # Skip the target if it's too far away from the viewpoint.
            target_distance = pixel_dist(ix_next_target, ix_viewpoint,
                                         iy_next_target, iy_viewpoint)*pixel_size
            if target_distance > max_visible_radius:
                continue

            # Checks pass, add the next target to the queue.
            ring_id = lmax(labs(ix_viewpoint_block - ix_next_target>>block_bits),
                           labs(iy_viewpoint_block - iy_next_target>>block_bits))
            process_queue.push(
                TargetPixel(ix_next_target, iy_next_target,
                            ring_id, target_pixel.sector, target_distance))
            process_queue_set.insert(next_target_index)
    LOGGER.info('%6.2f%% complete after %.2fs', 100.0, time.time()-start_time)

    dem_managed_raster.close()
    aux_managed_raster.close()
    visibility_managed_raster.close()

    if temp_dir is not None:
        try:
            shutil.rmtree(temp_dir)
        except OSError:
            LOGGER.exception('Could not remove temporary folder %s', temp_dir)



"""Module to that provides functions for usage logging."""
import contextlib
import hashlib
import json
import locale
import logging
import os
import platform
import sys
import threading
import traceback
import uuid
import importlib

from urllib.request import urlopen, Request
from urllib.parse import urlencode

from osgeo import osr
import natcap.invest
import pygeoprocessing

from . import utils

ENCODING = sys.getfilesystemencoding()
LOGGER = logging.getLogger(__name__)

_ENDPOINTS_INDEX_URL = (
    'http://data.naturalcapitalproject.org/server_registry/'
    'invest_usage_logger_v2/index.html')

# This is defined here because it's very useful to know the thread name ahead
# of time so we can exclude any log messages it generates from the logging.
# Python doesn't care about having multiple threads have the same name.
_USAGE_LOGGING_THREAD_NAME = 'usage-logging-thread'


@contextlib.contextmanager
def log_run(model_pyname, args):
    """Context manager to log an InVEST model run and exit status.

    Args:
        model_pyname (string): The string module name that identifies the model.
        args (dict): The full args dictionary.

    Returns:
        ``None``
    """
    invest_interface = 'Qt'  # this cm is only used by the Qt interface
    session_id = str(uuid.uuid4())
    log_thread = threading.Thread(
        target=_log_model,
        args=(model_pyname, args, invest_interface, session_id),
        name=_USAGE_LOGGING_THREAD_NAME)
    log_thread.start()

    try:
        yield
    except Exception:
        exit_status_message = traceback.format_exc()
        raise
    else:
        exit_status_message = ':)'
    finally:
        log_exit_thread = threading.Thread(
            target=_log_exit_status, args=(session_id, exit_status_message),
            name=_USAGE_LOGGING_THREAD_NAME)
        log_exit_thread.start()


def _calculate_args_bounding_box(args, args_spec):
    """Calculate the bounding boxes of any GIS types found in `args_dict`.

    Args:
        args (dict): a string key and any value pair dictionary.
        args_spec (dict): the model ARGS_SPEC describing args

    Returns:
        bb_intersection, bb_union tuple that's either the lat/lng bounding
            intersection and union bounding boxes of the gis types referred to
            in args_dict.  If no GIS types are present, this is a (None, None)
            tuple.
    """

    def _merge_bounding_boxes(bb1, bb2, mode):
        """Merge two bounding boxes through union or intersection.

        Args:
            bb1 (list of float): bounding box of the form
                [minx, maxy, maxx, miny] or None
            bb2 (list of float): bounding box of the form
                [minx, maxy, maxx, miny] or None
            mode (string): either "union" or "intersection" indicating the
                how to combine the two bounding boxes.

        Returns:
            either the intersection or union of bb1 and bb2 depending
            on mode.  If either bb1 or bb2 is None, the other is returned.
            If both are None, None is returned.
        """
        if bb1 is None:
            return bb2
        if bb2 is None:
            return bb1

        if mode == "union":
            comparison_ops = [min, max, max, min]
        if mode == "intersection":
            comparison_ops = [max, min, min, max]

        bb_out = [op(x, y) for op, x, y in zip(comparison_ops, bb1, bb2)]
        return bb_out

    bb_intersection = None
    bb_union = None
    for key, value in args.items():
        # Using gdal.OpenEx to check if an input is spatial caused the
        # model to hang sometimes (possible race condition), so only
        # get the bounding box of inputs that are known to be spatial.
        # Also eliminate any string paths that are empty to prevent an
        # exception. By the time we've made it to this function, all paths
        # should already have been validated so the path is either valid or
        # blank.
        spatial_info = None
        if args_spec['args'][key]['type'] == 'raster' and value.strip() != '':
            spatial_info = pygeoprocessing.get_raster_info(value)
        elif (args_spec['args'][key]['type'] == 'vector'
                and value.strip() != ''):
            spatial_info = pygeoprocessing.get_vector_info(value)

        if spatial_info:
            local_bb = spatial_info['bounding_box']
            projection_wkt = spatial_info['projection_wkt']
            spatial_ref = osr.SpatialReference()
            spatial_ref.ImportFromWkt(projection_wkt)

            try:
                # means there's a GIS type with a well defined bounding box
                # create transform, and reproject local bounding box to
                # lat/lng
                lat_lng_ref = osr.SpatialReference()
                lat_lng_ref.ImportFromEPSG(4326)  # EPSG 4326 is lat/lng
                to_lat_trans = utils.create_coordinate_transformer(
                    spatial_ref, lat_lng_ref)
                for point_index in [0, 2]:
                    local_bb[point_index], local_bb[point_index + 1], _ = (
                        to_lat_trans.TransformPoint(
                            local_bb[point_index],
                            local_bb[point_index+1]))

                bb_intersection = _merge_bounding_boxes(
                    local_bb, bb_intersection, 'intersection')
                bb_union = _merge_bounding_boxes(
                    local_bb, bb_union, 'union')
            except Exception as transform_error:
                # All kinds of exceptions from bad transforms or CSV files
                # or dbf files could get us to this point, just don't
                # bother with the local_bb at all
                LOGGER.exception('Error when transforming coordinates: %s',
                                 transform_error)
        else:
            LOGGER.debug(f'Arg {key} of type {args_spec["args"][key]["type"]} '
                          'excluded from bounding box calculation')

    return bb_intersection, bb_union


def _log_exit_status(session_id, status):
    """Log the completion of a model with the given status.

    Args:
        session_id (string): a unique string that can be used to identify
            the current session between the model initial start and exit.
        status (string): a string describing the exit status of the model,
            'success' would indicate the successful completion while an
            exception string could indicate a failure.

    Returns:
        None
    """
    logger = logging.getLogger('natcap.invest.ui.usage._log_exit_status')

    try:
        payload = {
            'session_id': session_id,
            'status': status,
        }
        log_finish_url = json.loads(urlopen(
            _ENDPOINTS_INDEX_URL).read().strip())['FINISH']

        # The data must be a python string of bytes. This will be ``bytes``
        # in python3.
        urlopen(Request(log_finish_url, urlencode(payload).encode('utf-8')))
    except Exception as exception:
        # An exception was thrown, we don't care.
        logger.warning(
            'an exception encountered when _log_exit_status %s',
            str(exception))


def _log_model(pyname, model_args, invest_interface, session_id=None):
    """Log information about a model run to a remote server.

    Args:
        pyname (string): a python string of the package version.
        model_args (dict): the traditional InVEST argument dictionary.
        invest_interface (string): a string identifying the calling UI,
            e.g. `Qt` or 'Workbench'.

    Returns:
        None
    """
    logger = logging.getLogger('natcap.invest.ui.usage._log_model')

    def _node_hash():
        """Return a hash for the current computational node."""
        data = {
            'os': platform.platform(),
            'hostname': platform.node(),
            'userdir': os.path.expanduser('~')
        }
        md5 = hashlib.md5()
        # a json dump will handle non-ascii encodings
        # but then data must be encoded before hashing in Python 3.
        md5.update(json.dumps(data).encode('utf-8'))
        return md5.hexdigest()

    args_spec = importlib.import_module(pyname).ARGS_SPEC

    try:
        bounding_box_intersection, bounding_box_union = (
            _calculate_args_bounding_box(model_args, args_spec))

        payload = {
            'model_name': pyname,
            'invest_release': natcap.invest.__version__,
            'invest_interface': invest_interface,
            'node_hash': _node_hash(),
            'system_full_platform_string': platform.platform(),
            'system_preferred_encoding': locale.getdefaultlocale()[1],
            'system_default_language': locale.getdefaultlocale()[0],
            'bounding_box_intersection': str(bounding_box_intersection),
            'bounding_box_union': str(bounding_box_union),
            'session_id': session_id,
        }
        log_start_url = json.loads(urlopen(
            _ENDPOINTS_INDEX_URL).read().strip())['START']

        # The data must be a python string of bytes. This will be ``bytes``
        # in python3.
        urlopen(Request(log_start_url, urlencode(payload).encode('utf-8')))
    except Exception as exception:
        # An exception was thrown, we don't care.
        logger.warning(
            'an exception encountered when logging %s', repr(exception))



# coding=UTF-8

from natcap.invest.ui import model, inputs
from natcap.invest import annual_water_yield, MODEL_METADATA


class AnnualWaterYield(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['annual_water_yield'].model_title,
            target=annual_water_yield.execute,
            validator=annual_water_yield.validate,
            localdoc=MODEL_METADATA['annual_water_yield'].userguide)

        self.precipitation = inputs.File(
            args_key='precipitation_path',
            helptext=(
                "A GDAL-supported raster file containing non-zero, "
                "average annual precipitation values for each cell. "
                "The precipitation values should be in millimeters "
                "(mm)."),
            label='Precipitation (Raster)',
            validator=self.validator)
        self.add_input(self.precipitation)
        self.potential_evapotranspiration = inputs.File(
            args_key='eto_path',
            helptext=(
                "A GDAL-supported raster file containing annual "
                "average reference evapotranspiration values for each "
                "cell.  The reference evapotranspiration values should "
                "be in millimeters (mm)."),
            label='Reference Evapotranspiration (Raster)',
            validator=self.validator)
        self.add_input(self.potential_evapotranspiration)
        self.depth_to_root_rest_layer = inputs.File(
            args_key='depth_to_root_rest_layer_path',
            helptext=(
                "A GDAL-supported raster file containing an average "
                "root restricting layer depth value for each cell. "
                "The root restricting layer depth value should be in "
                "millimeters (mm)."),
            label='Depth To Root Restricting Layer (Raster)',
            validator=self.validator)
        self.add_input(self.depth_to_root_rest_layer)
        self.plant_available_water_fraction = inputs.File(
            args_key='pawc_path',
            helptext=(
                "A GDAL-supported raster file containing plant "
                "available water content values for each cell.  The "
                "plant available water content fraction should be a "
                "value between 0 and 1."),
            label='Plant Available Water Fraction (Raster)',
            validator=self.validator)
        self.add_input(self.plant_available_water_fraction)
        self.land_use = inputs.File(
            args_key='lulc_path',
            helptext=(
                "A GDAL-supported raster file containing LULC code "
                "(expressed as integers) for each cell."),
            label='Land Use (Raster)',
            validator=self.validator)
        self.add_input(self.land_use)
        self.watersheds = inputs.File(
            args_key='watersheds_path',
            helptext=(
                "An OGR-supported vector file containing one polygon "
                "per watershed.  Each polygon that represents a "
                "watershed is required to have a field 'ws_id' that is "
                "a unique integer which identifies that watershed."),
            label='Watersheds (Vector)',
            validator=self.validator)
        self.add_input(self.watersheds)
        self.sub_watersheds = inputs.File(
            args_key='sub_watersheds_path',
            helptext=(
                "An OGR-supported vector file with one polygon per "
                "sub-watershed within the main watersheds specified in "
                "the Watersheds shapefile.  Each polygon that "
                "represents a sub-watershed is required to have a "
                "field 'subws_id' that is a unique integer which "
                "identifies that sub-watershed."),
            label='Sub-Watersheds (Vector) (Optional)',
            validator=self.validator)
        self.add_input(self.sub_watersheds)
        self.biophysical_table = inputs.File(
            args_key='biophysical_table_path',
            helptext=(
                "A CSV table of land use/land cover (LULC) classes, "
                "containing data on biophysical coefficients used in "
                "this model.  The following columns are required: "
                "'lucode' (integer), 'root_depth' (mm), 'Kc' "
                "(coefficient)."),
            label='Biophysical Table (CSV)',
            validator=self.validator)
        self.add_input(self.biophysical_table)
        self.seasonality_constant = inputs.Text(
            args_key='seasonality_constant',
            helptext=(
                "Floating point value on the order of 1 to 30 "
                "corresponding to the seasonal distribution of "
                "precipitation."),
            label='Z parameter',
            validator=self.validator)
        self.add_input(self.seasonality_constant)
        self.scarcity_valuation_container = inputs.Container(
            args_key='do_scarcity_and_valuation',
            expandable=True,
            expanded=False,
            label='Water Scarcity and Valuation')
        self.add_input(self.scarcity_valuation_container)
        self.demand_table = inputs.File(
            args_key='demand_table_path',
            helptext=(
                "A CSV table of LULC classes, showing consumptive "
                "water use for each land-use/land-cover type.  The "
                "table requires two column fields: 'lucode' and "
                "'demand'. The demand values should be the estimated "
                "average consumptive water use for each land-use/land- "
                "cover type.  Water use should be given in cubic "
                "meters per year for a pixel in the land-use/land- "
                "cover map.  NOTE: the accounting for pixel area is "
                "important since larger areas will consume more water "
                "for the same land-cover type."),
            label='Water Demand Table (CSV)',
            validator=self.validator)
        self.scarcity_valuation_container.add_input(self.demand_table)
        self.hydropower_valuation_table = inputs.File(
            args_key='valuation_table_path',
            helptext=(
                "A CSV table of hydropower stations with associated "
                "model values.  The table should have the following "
                "column fields: 'ws_id', 'efficiency', 'fraction', "
                "'height', 'kw_price', 'cost', 'time_span', and "
                "'discount'."),
            label='Hydropower Valuation Table (CSV) (Optional)',
            validator=self.validator)
        self.scarcity_valuation_container.add_input(self.hydropower_valuation_table)

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.precipitation.args_key: self.precipitation.value(),
            self.potential_evapotranspiration.args_key:
                self.potential_evapotranspiration.value(),
            self.depth_to_root_rest_layer.args_key:
                self.depth_to_root_rest_layer.value(),
            self.plant_available_water_fraction.args_key:
                self.plant_available_water_fraction.value(),
            self.land_use.args_key: self.land_use.value(),
            self.watersheds.args_key: self.watersheds.value(),
            self.sub_watersheds.args_key: self.sub_watersheds.value(),
            self.biophysical_table.args_key: self.biophysical_table.value(),
            self.seasonality_constant.args_key:
                self.seasonality_constant.value(),
            self.scarcity_valuation_container.args_key:
                self.scarcity_valuation_container.value(),
        }

        if self.scarcity_valuation_container.value():
            args[self.demand_table.args_key] = self.demand_table.value()
            args[self.hydropower_valuation_table.args_key] = (
                self.hydropower_valuation_table.value())

        return args



"""Stormwater Retention."""
import logging
import math
import os

import numpy
from osgeo import gdal, ogr, osr
import pygeoprocessing
import taskgraph

from . import spec_utils
from .spec_utils import u
from . import utils
from . import validation
from . import MODEL_METADATA


LOGGER = logging.getLogger(__name__)

# a constant nodata value to use for intermediates and outputs
FLOAT_NODATA = -1
UINT8_NODATA = 255
UINT16_NODATA = 65535

ARGS_SPEC = {
    "model_name": MODEL_METADATA["stormwater"].model_title,
    "pyname": MODEL_METADATA["stormwater"].pyname,
    "userguide_html": MODEL_METADATA["stormwater"].userguide,
    "args_with_spatial_overlap": {
        "spatial_keys": ["lulc_path", "soil_group_path", "precipitation_path",
                         "road_centerlines_path", "aggregate_areas_path"],
        "different_projections_ok": True
    },
    "args": {
        "workspace_dir": spec_utils.WORKSPACE,
        "results_suffix": spec_utils.SUFFIX,
        "n_workers": spec_utils.N_WORKERS,
        "lulc_path": {
            **spec_utils.LULC,
            "projected": True
        },
        "soil_group_path": spec_utils.SOIL_GROUP,
        "precipitation_path": spec_utils.PRECIP,
        "biophysical_table": {
            "type": "csv",
            "columns": {
                "lucode": {
                    "type": "integer",
                    "about": _("LULC code corresponding to the LULC raster")
                },
                "emc_[POLLUTANT]": {
                    "type": "number",
                    "units": u.milligram/u.liter,
                    "about": _(
                        "Event mean concentration of the pollutant in "
                        "stormwater. You may include any number of these "
                        "columns for different pollutants, or none at all.")
                },
                **{
                    f"rc_{soil_group}": {
                        "type": "ratio",
                        "about": _("Stormwater runoff coefficient for soil "
                                  f"group {soil_group.upper()}")
                    } for soil_group in ["a", "b", "c", "d"]
                },
                **{
                    f"pe_{soil_group}": {
                        "type": "ratio",
                        "about": _("Stormwater percolation coefficient for "
                                  f"soil group {soil_group.upper()}"),
                        "required": False
                    } for soil_group in ["a", "b", "c", "d"]
                },
                "is_connected": {
                    "type": "boolean",
                    "required": False,
                    "about": _(
                        "Enter 1 if the LULC class is a connected impervious "
                        "surface, 0 if not. This column is only used if the "
                        "'adjust retention ratios' option is selected. If "
                        "'adjust retention ratios' is selected and this "
                        "column exists, the adjustment algorithm takes into "
                        "account the LULC as well as road centerlines. If "
                        "this column does not exist, only the road "
                        "centerlines are used.")
                }
            },
            "about": _(
                "Table mapping each LULC code found in the LULC raster to "
                "biophysical data about that LULC class. If you provide the "
                "percolation coefficient column (PE_[X]) for any soil group, "
                "you must provide it for all four soil groups."),
            "name": _("Biophysical table")
        },
        "adjust_retention_ratios": {
            "type": "boolean",
            "about": _(
                "If true, adjust retention ratios. The adjustment algorithm "
                "accounts for drainage effects of nearby impervious surfaces "
                "which are directly connected to artifical urban drainage "
                "channels (typically roads, parking lots, etc.) Connected "
                "impervious surfaces are indicated by the is_connected column"
                "in the biophysical table and/or the road centerlines vector."),
            "name": _("Adjust retention ratios")
        },
        "retention_radius": {
            "type": "number",
            "units": u.linear_unit,
            "required": "adjust_retention_ratios",
            "about": _(
                "Radius around each pixel to adjust retention ratios. "
                "Measured in raster coordinate system units. For the "
                "adjustment algorithm, a pixel is 'near' a connected "
                "impervious surface if its centerpoint is within this radius "
                "of connected-impervious LULC and/or a road centerline."),
            "name": _("Retention radius")
        },
        "road_centerlines_path": {
            "type": "vector",
            "geometries": {"LINESTRING", "MULTILINESTRING"},
            "fields": {},
            "required": "adjust_retention_ratios",
            "about": _("Map of road centerlines"),
            "name": _("Road centerlines")
        },
        "aggregate_areas_path": {
            **spec_utils.AOI,
            "required": False,
            "about": _(
                "Areas over which to aggregate results (typically watersheds "
                "or sewersheds). The aggregated data are: average retention "
                "ratio and total retention volume; average percolation ratio "
                "and total percolation volume if percolation data was "
                "provided; total retention value if replacement cost was "
                "provided; and total avoided pollutant load for each "
                "pollutant provided."),
        },
        "replacement_cost": {
            "type": "number",
            "units": u.currency/u.meter**3,
            "required": False,
            "about": _("Replacement cost of stormwater retention devices"),
            "name": _("Replacement cost")
        }
    }
}

INTERMEDIATE_OUTPUTS = {
    'lulc_aligned_path': 'lulc_aligned.tif',
    'soil_group_aligned_path': 'soil_group_aligned.tif',
    'precipitation_aligned_path': 'precipitation_aligned.tif',
    'reprojected_centerlines_path': 'reprojected_centerlines.gpkg',
    'rasterized_centerlines_path': 'rasterized_centerlines.tif',
    'connected_lulc_path': 'is_connected_lulc.tif',
    'road_distance_path': 'road_distance.tif',
    'search_kernel_path': 'search_kernel.tif',
    'connected_lulc_distance_path': 'connected_lulc_distance.tif',
    'near_connected_lulc_path': 'near_connected_lulc.tif',
    'near_road_path': 'near_road.tif',
    'ratio_average_path': 'ratio_average.tif'
}

FINAL_OUTPUTS = {
    'reprojected_aggregate_areas_path': 'aggregate_data.gpkg',
    'retention_ratio_path': 'retention_ratio.tif',
    'adjusted_retention_ratio_path': 'adjusted_retention_ratio.tif',
    'runoff_ratio_path': 'runoff_ratio.tif',
    'retention_volume_path': 'retention_volume.tif',
    'runoff_volume_path': 'runoff_volume.tif',
    'percolation_ratio_path': 'percolation_ratio.tif',
    'percolation_volume_path': 'percolation_volume.tif',
    'retention_value_path': 'retention_value.tif'
}


def execute(args):
    """Execute the urban stormwater retention model.

    Args:
        args['workspace_dir'] (str): path to a directory to write intermediate
            and final outputs. May already exist or not.
        args['results_suffix'] (str, optional): string to append to all output
            file names from this model run
        args['n_workers'] (int): if present, indicates how many worker
            processes should be used in parallel processing. -1 indicates
            single process mode, 0 is single process but non-blocking mode,
            and >= 1 is number of processes.
        args['lulc_path'] (str): path to LULC raster
        args['soil_group_path'] (str): path to soil group raster, where pixel
            values 1, 2, 3, 4 correspond to groups A, B, C, D
        args['precipitation_path'] (str): path to raster of total annual
            precipitation in millimeters
        args['biophysical_table'] (str): path to biophysical table with columns
            'lucode', 'EMC_x' (event mean concentration mg/L) for each
            pollutant x, 'RC_y' (retention coefficient) and 'PE_y'
            (percolation coefficient) for each soil group y, and
            'is_connected' if args['adjust_retention_ratios'] is True
        args['adjust_retention_ratios'] (bool): If True, apply retention ratio
            adjustment algorithm.
        args['retention_radius'] (float): If args['adjust_retention_ratios']
            is True, use this radius in the adjustment algorithm.
        args['road_centerliens_path'] (str): Path to linestring vector of road
            centerlines. Only used if args['adjust_retention_ratios'] is True.
        args['aggregate_areas_path'] (str): Optional path to polygon vector of
            areas to aggregate results over.
        args['replacement_cost'] (float): Cost to replace stormwater retention
            devices in units currency per cubic meter

    Returns:
        None
    """
    # set up files and directories
    suffix = utils.make_suffix_string(args, 'results_suffix')
    output_dir = args['workspace_dir']
    intermediate_dir = os.path.join(output_dir, 'intermediate')
    cache_dir = os.path.join(intermediate_dir, 'cache_dir')
    utils.make_directories(
        [args['workspace_dir'], intermediate_dir, cache_dir])
    files = utils.build_file_registry(
        [(INTERMEDIATE_OUTPUTS, intermediate_dir),
         (FINAL_OUTPUTS, output_dir)], suffix)

    task_graph = taskgraph.TaskGraph(cache_dir, int(args.get('n_workers', -1)))

    # get the necessary base raster info
    source_lulc_raster_info = pygeoprocessing.get_raster_info(
        args['lulc_path'])
    pixel_size = source_lulc_raster_info['pixel_size']
    # in case the input raster doesn't have square pixels, take the average
    # all the rasters are warped to this square pixel size in the align task
    avg_pixel_size = (abs(pixel_size[0]) + abs(pixel_size[1])) / 2
    pixel_area = abs(pixel_size[0] * pixel_size[1])

    lulc_nodata = source_lulc_raster_info['nodata'][0]
    precipitation_nodata = pygeoprocessing.get_raster_info(
        args['precipitation_path'])['nodata'][0]
    soil_group_nodata = pygeoprocessing.get_raster_info(
        args['soil_group_path'])['nodata'][0]

    # Align all three input rasters to the same projection
    align_inputs = [args['lulc_path'],
                    args['soil_group_path'], args['precipitation_path']]
    align_outputs = [
        files['lulc_aligned_path'],
        files['soil_group_aligned_path'],
        files['precipitation_aligned_path']]
    align_task = task_graph.add_task(
        func=pygeoprocessing.align_and_resize_raster_stack,
        args=(
            align_inputs,
            align_outputs,
            ['near' for _ in align_inputs],
            (avg_pixel_size, -avg_pixel_size),
            'intersection'),
        kwargs={'raster_align_index': 0},
        target_path_list=align_outputs,
        task_name='align input rasters')

    # Build a lookup dictionary mapping each LULC code to its row
    biophysical_dict = utils.build_lookup_from_csv(
        args['biophysical_table'], 'lucode')
    # sort the LULC codes upfront because we use the sorted list in multiple
    # places. it's more efficient to do this once.
    sorted_lucodes = sorted(biophysical_dict)

    # convert the nested dictionary in to a 2D array where rows are LULC codes
    # in sorted order and columns correspond to soil groups in order
    # this facilitates efficiently looking up the ratio values with numpy
    #
    # Biophysical table has runoff coefficents so subtract
    # from 1 to get retention coefficient.
    # add a placeholder in column 0 so that the soil groups 1, 2, 3, 4 line
    # up with their indices in the array. this is more efficient than
    # decrementing the whole soil group array by 1.
    retention_ratio_array = numpy.array([
        [1 - biophysical_dict[lucode][f'rc_{soil_group}']
            for soil_group in ['a', 'b', 'c', 'd']
         ] for lucode in sorted_lucodes
    ], dtype=numpy.float32)

    # Calculate stormwater retention ratio and volume from
    # LULC, soil groups, biophysical data, and precipitation
    retention_ratio_task = task_graph.add_task(
        func=lookup_ratios,
        args=(
            files['lulc_aligned_path'],
            files['soil_group_aligned_path'],
            retention_ratio_array,
            sorted_lucodes,
            files['retention_ratio_path']),
        target_path_list=[files['retention_ratio_path']],
        dependent_task_list=[align_task],
        task_name='calculate stormwater retention ratio'
    )

    # (Optional) adjust stormwater retention ratio using roads
    if args['adjust_retention_ratios']:
        # in raster coord system units
        radius = float(args['retention_radius'])
        # boolean mapping for each LULC code whether it's connected
        is_connected_map = {
            lucode: 1 if biophysical_dict[lucode]['is_connected'] else 0
            for lucode in biophysical_dict}

        reproject_roads_task = task_graph.add_task(
            func=pygeoprocessing.reproject_vector,
            args=(
                args['road_centerlines_path'],
                source_lulc_raster_info['projection_wkt'],
                files['reprojected_centerlines_path']),
            kwargs={'driver_name': 'GPKG'},
            target_path_list=[files['reprojected_centerlines_path']],
            task_name='reproject road centerlines vector to match rasters',
            dependent_task_list=[])

        # for gdal.GDT_Byte, setting the datatype is not enough
        # must also set PIXELTYPE=DEFAULT to guarantee unsigned byte type
        # otherwise, `new_raster_from_base` may pass down the
        # PIXELTYPE=SIGNEDBYTE attribute from a signed base to the new raster
        creation_opts = pygeoprocessing.geoprocessing_core.DEFAULT_GTIFF_CREATION_TUPLE_OPTIONS[1]  # noqa
        unsigned_byte_creation_opts = creation_opts + ('PIXELTYPE=DEFAULT',)
        # pygeoprocessing.rasterize expects the target raster to already exist
        make_raster_task = task_graph.add_task(
            func=pygeoprocessing.new_raster_from_base,
            args=(
                files['lulc_aligned_path'],
                files['rasterized_centerlines_path'],
                gdal.GDT_Byte,
                [UINT8_NODATA]),
            kwargs={
                'raster_driver_creation_tuple': (
                    'GTIFF', unsigned_byte_creation_opts)
            },
            target_path_list=[files['rasterized_centerlines_path']],
            task_name='create raster to pass to pygeoprocessing.rasterize',
            dependent_task_list=[align_task])

        rasterize_centerlines_task = task_graph.add_task(
            func=pygeoprocessing.rasterize,
            args=(
                files['reprojected_centerlines_path'],
                files['rasterized_centerlines_path'],
                [1]),
            target_path_list=[files['rasterized_centerlines_path']],
            task_name='rasterize road centerlines vector',
            dependent_task_list=[make_raster_task, reproject_roads_task])

        # Make a boolean raster showing which pixels are within the given
        # radius of a road centerline
        near_road_task = task_graph.add_task(
            func=is_near,
            args=(
                files['rasterized_centerlines_path'],
                radius / avg_pixel_size,  # convert the radius to pixels
                files['road_distance_path'],
                files['near_road_path']),
            target_path_list=[
                files['road_distance_path'],
                files['near_road_path']],
            task_name='find pixels within radius of road centerlines',
            dependent_task_list=[rasterize_centerlines_task])

        # Make a boolean raster indicating which pixels are directly
        # connected impervious LULC type
        connected_lulc_task = task_graph.add_task(
            func=pygeoprocessing.reclassify_raster,
            args=(
                (files['lulc_aligned_path'], 1),
                is_connected_map,
                files['connected_lulc_path'],
                gdal.GDT_Byte,
                UINT8_NODATA),
            target_path_list=[files['connected_lulc_path']],
            task_name='calculate binary connected lulc raster',
            dependent_task_list=[align_task]
        )

        # Make a boolean raster showing which pixels are within the given
        # radius of connected land cover
        near_connected_lulc_task = task_graph.add_task(
            func=is_near,
            args=(
                files['connected_lulc_path'],
                radius / avg_pixel_size,  # convert the radius to pixels
                files['connected_lulc_distance_path'],
                files['near_connected_lulc_path']),
            target_path_list=[
                files['connected_lulc_distance_path'],
                files['near_connected_lulc_path']],
            task_name='find pixels within radius of connected lulc',
            dependent_task_list=[connected_lulc_task])

        average_ratios_task = task_graph.add_task(
            func=raster_average,
            args=(
                files['retention_ratio_path'],
                radius,
                files['search_kernel_path'],
                files['ratio_average_path']),
            target_path_list=[files['ratio_average_path']],
            task_name='average retention ratios within radius',
            dependent_task_list=[retention_ratio_task])

        # Using the averaged retention ratio raster and boolean
        # "within radius" rasters, adjust the retention ratios
        adjust_retention_ratio_task = task_graph.add_task(
            func=pygeoprocessing.raster_calculator,
            args=([
                (files['retention_ratio_path'], 1),
                (files['ratio_average_path'], 1),
                (files['near_connected_lulc_path'], 1),
                (files['near_road_path'], 1)],
                adjust_op,
                files['adjusted_retention_ratio_path'],
                gdal.GDT_Float32,
                FLOAT_NODATA),
            target_path_list=[files['adjusted_retention_ratio_path']],
            task_name='adjust stormwater retention ratio',
            dependent_task_list=[retention_ratio_task, average_ratios_task,
                                 near_connected_lulc_task, near_road_task])

        final_retention_ratio_path = files['adjusted_retention_ratio_path']
        final_retention_ratio_task = adjust_retention_ratio_task
    else:
        final_retention_ratio_path = files['retention_ratio_path']
        final_retention_ratio_task = retention_ratio_task

    # Calculate stormwater retention volume from ratios and precipitation
    retention_volume_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=([
            (final_retention_ratio_path, 1),
            (files['precipitation_aligned_path'], 1),
            (precipitation_nodata, 'raw'),
            (pixel_area, 'raw')],
            volume_op,
            files['retention_volume_path'],
            gdal.GDT_Float32,
            FLOAT_NODATA),
        target_path_list=[files['retention_volume_path']],
        dependent_task_list=[align_task, final_retention_ratio_task],
        task_name='calculate stormwater retention volume'
    )

    # Calculate stormwater runoff ratios and volume
    runoff_ratio_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=(
            [(final_retention_ratio_path, 1)],
            retention_to_runoff_op,
            files['runoff_ratio_path'],
            gdal.GDT_Float32,
            FLOAT_NODATA),
        target_path_list=[files['runoff_ratio_path']],
        dependent_task_list=[final_retention_ratio_task],
        task_name='calculate stormwater runoff ratio'
    )

    runoff_volume_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=([
            (files['runoff_ratio_path'], 1),
            (files['precipitation_aligned_path'], 1),
            (precipitation_nodata, 'raw'),
            (pixel_area, 'raw')],
            volume_op,
            files['runoff_volume_path'],
            gdal.GDT_Float32,
            FLOAT_NODATA),
        target_path_list=[files['runoff_volume_path']],
        dependent_task_list=[align_task, runoff_ratio_task],
        task_name='calculate stormwater runoff volume'
    )
    aggregation_task_dependencies = [retention_volume_task]
    data_to_aggregate = [
        # tuple of (raster path, output field name, op) for aggregation
        (final_retention_ratio_path, 'mean_retention_ratio', 'mean'),
        (files['retention_volume_path'], 'total_retention_volume', 'sum'),
        (files['runoff_ratio_path'], 'mean_runoff_ratio', 'mean'),
        (files['runoff_volume_path'], 'total_runoff_volume', 'sum')]

    # (Optional) Calculate stormwater percolation ratio and volume from
    # LULC, soil groups, biophysical table, and precipitation
    if 'pe_a' in next(iter(biophysical_dict.values())):
        LOGGER.info('percolation data detected in biophysical table. '
                    'Will calculate percolation ratio and volume rasters.')
        percolation_ratio_array = numpy.array([
            [biophysical_dict[lucode][f'pe_{soil_group}']
                for soil_group in ['a', 'b', 'c', 'd']
             ] for lucode in sorted_lucodes
        ], dtype=numpy.float32)
        percolation_ratio_task = task_graph.add_task(
            func=lookup_ratios,
            args=(
                files['lulc_aligned_path'],
                files['soil_group_aligned_path'],
                percolation_ratio_array,
                sorted_lucodes,
                files['percolation_ratio_path']),
            target_path_list=[files['percolation_ratio_path']],
            dependent_task_list=[align_task],
            task_name='calculate stormwater percolation ratio')

        percolation_volume_task = task_graph.add_task(
            func=pygeoprocessing.raster_calculator,
            args=([
                (files['percolation_ratio_path'], 1),
                (files['precipitation_aligned_path'], 1),
                (precipitation_nodata, 'raw'),
                (pixel_area, 'raw')],
                volume_op,
                files['percolation_volume_path'],
                gdal.GDT_Float32,
                FLOAT_NODATA),
            target_path_list=[files['percolation_volume_path']],
            dependent_task_list=[align_task, percolation_ratio_task],
            task_name='calculate stormwater retention volume'
        )
        aggregation_task_dependencies.append(percolation_volume_task)
        data_to_aggregate.append((files['percolation_ratio_path'],
                                 'mean_percolation_ratio', 'mean'))
        data_to_aggregate.append((files['percolation_volume_path'],
                                 'total_percolation_volume', 'sum'))

    # get all EMC columns from an arbitrary row in the dictionary
    # strip the first four characters off 'EMC_pollutant' to get pollutant name
    pollutants = [key[4:] for key in next(iter(biophysical_dict.values()))
                  if key.startswith('emc_')]
    LOGGER.debug(f'Pollutants found in biophysical table: {pollutants}')

    # Calculate avoided pollutant load for each pollutant from retention volume
    # and biophysical table EMC value
    avoided_load_paths = []
    actual_load_paths = []
    for pollutant in pollutants:
        # two output rasters for each pollutant
        avoided_pollutant_load_path = os.path.join(
            output_dir, f'avoided_pollutant_load_{pollutant}{suffix}.tif')
        avoided_load_paths.append(avoided_pollutant_load_path)
        actual_pollutant_load_path = os.path.join(
            output_dir, f'actual_pollutant_load_{pollutant}{suffix}.tif')
        actual_load_paths.append(actual_pollutant_load_path)
        # make an array mapping each LULC code to the pollutant EMC value
        emc_array = numpy.array(
            [biophysical_dict[lucode][f'emc_{pollutant}']
                for lucode in sorted_lucodes], dtype=numpy.float32)

        # calculate avoided load from retention volume
        avoided_load_task = task_graph.add_task(
            func=pygeoprocessing.raster_calculator,
            args=([
                (files['lulc_aligned_path'], 1),
                (lulc_nodata, 'raw'),
                (files['retention_volume_path'], 1),
                (sorted_lucodes, 'raw'),
                (emc_array, 'raw')],
                pollutant_load_op,
                avoided_pollutant_load_path,
                gdal.GDT_Float32,
                FLOAT_NODATA),
            target_path_list=[avoided_pollutant_load_path],
            dependent_task_list=[retention_volume_task],
            task_name=f'calculate avoided pollutant {pollutant} load'
        )
        # calculate actual load from runoff volume
        actual_load_task = task_graph.add_task(
            func=pygeoprocessing.raster_calculator,
            args=([
                (files['lulc_aligned_path'], 1),
                (lulc_nodata, 'raw'),
                (files['runoff_volume_path'], 1),
                (sorted_lucodes, 'raw'),
                (emc_array, 'raw')],
                pollutant_load_op,
                actual_pollutant_load_path,
                gdal.GDT_Float32,
                FLOAT_NODATA),
            target_path_list=[actual_pollutant_load_path],
            dependent_task_list=[runoff_volume_task],
            task_name=f'calculate actual pollutant {pollutant} load'
        )
        aggregation_task_dependencies += [avoided_load_task, actual_load_task]
        data_to_aggregate.append((avoided_pollutant_load_path,
                                 f'{pollutant}_total_avoided_load', 'sum'))
        data_to_aggregate.append(
            (actual_pollutant_load_path, f'{pollutant}_total_load', 'sum'))

    # (Optional) Do valuation if a replacement cost is defined
    # you could theoretically have a cost of 0 which should be allowed
    if 'replacement_cost' in args and args['replacement_cost'] not in [
            None, '']:
        valuation_task = task_graph.add_task(
            func=pygeoprocessing.raster_calculator,
            args=([
                (files['retention_volume_path'], 1),
                (float(args['replacement_cost']), 'raw')],
                retention_value_op,
                files['retention_value_path'],
                gdal.GDT_Float32,
                FLOAT_NODATA),
            target_path_list=[files['retention_value_path']],
            dependent_task_list=[retention_volume_task],
            task_name='calculate stormwater retention value'
        )
        aggregation_task_dependencies.append(valuation_task)
        data_to_aggregate.append(
            (files['retention_value_path'], 'total_retention_value', 'sum'))

    # (Optional) Aggregate to watersheds if an aggregate vector is defined
    if 'aggregate_areas_path' in args and args['aggregate_areas_path']:
        _ = task_graph.add_task(
            func=aggregate_results,
            args=(
                args['aggregate_areas_path'],
                files['reprojected_aggregate_areas_path'],
                source_lulc_raster_info['projection_wkt'],
                data_to_aggregate),
            target_path_list=[files['reprojected_aggregate_areas_path']],
            dependent_task_list=aggregation_task_dependencies,
            task_name='aggregate data over polygons'
        )

    task_graph.close()
    task_graph.join()


def lookup_ratios(lulc_path, soil_group_path, ratio_lookup, sorted_lucodes,
                  output_path):
    """Look up retention/percolation ratios from LULC codes and soil groups.

    Args:
        lulc_array (numpy.ndarray): 2D array of LULC codes
        soil_group_array (numpy.ndarray): 2D array with the same shape as
            ``lulc_array``. Values in {1, 2, 3, 4} corresponding to soil
            groups A, B, C, and D.
        ratio_lookup (numpy.ndarray): 2D array where rows correspond to
            sorted LULC codes and the 4 columns correspond to soil groups
            A, B, C, D in order. Shape: (number of lulc codes, 4).
        sorted_lucodes (list[int]): List of LULC codes sorted from smallest
            to largest. These correspond to the rows of ``ratio_lookup``.
        output_path (str): path to a raster to write out the result. has the
            same shape as the lulc and soil group rasters. Each value is the
            corresponding ratio for that LULC code x soil group pair.

    Returns:
        None
    """
    lulc_nodata = pygeoprocessing.get_raster_info(lulc_path)['nodata'][0]
    soil_group_nodata = pygeoprocessing.get_raster_info(
        soil_group_path)['nodata'][0]
    # insert a column on the left side of the array so that the soil
    # group codes 1-4 line up with their indexes. this is faster than
    # decrementing every value in a large raster.
    ratio_lookup = numpy.insert(ratio_lookup, 0,
                                numpy.zeros(ratio_lookup.shape[0]), axis=1)

    def ratio_op(lulc_array, soil_group_array):
        output_ratio_array = numpy.full(lulc_array.shape, FLOAT_NODATA,
                                        dtype=numpy.float32)
        valid_mask = numpy.full(lulc_array.shape, True)
        if lulc_nodata is not None:
            valid_mask &= ~utils.array_equals_nodata(lulc_array, lulc_nodata)
        if soil_group_nodata is not None:
            valid_mask &= ~utils.array_equals_nodata(
                soil_group_array, soil_group_nodata)
        # the index of each lucode in the sorted lucodes array
        lulc_index = numpy.digitize(lulc_array[valid_mask], sorted_lucodes,
                                    right=True)
        output_ratio_array[valid_mask] = (
            ratio_lookup[lulc_index, soil_group_array[valid_mask]])
        return output_ratio_array

    pygeoprocessing.raster_calculator(
        [(lulc_path, 1), (soil_group_path, 1)],
        ratio_op,
        output_path,
        gdal.GDT_Float32,
        FLOAT_NODATA)


def volume_op(ratio_array, precip_array, precip_nodata, pixel_area):
    """Calculate stormwater volumes from precipitation and ratios.

    Args:
        ratio_array (numpy.ndarray): 2D array of stormwater ratios. Assuming
            that its nodata value is the global FLOAT_NODATA.
        precip_array (numpy.ndarray): 2D array of precipitation amounts
            in millimeters/year
        precip_nodata (float): nodata value for the precipitation array
        pixel_area (float): area of each pixel in m^2

    Returns:
        2D numpy.ndarray of precipitation volumes in m^3/year
    """
    volume_array = numpy.full(ratio_array.shape, FLOAT_NODATA,
                              dtype=numpy.float32)
    valid_mask = ~utils.array_equals_nodata(ratio_array, FLOAT_NODATA)
    if precip_nodata is not None:
        valid_mask &= ~utils.array_equals_nodata(precip_array, precip_nodata)

    # precipitation (mm/yr) * pixel area (m^2) *
    # 0.001 (m/mm) * ratio = volume (m^3/yr)
    volume_array[valid_mask] = (
        precip_array[valid_mask] *
        ratio_array[valid_mask] *
        pixel_area * 0.001)
    return volume_array


def retention_to_runoff_op(retention_array):
    """Calculate runoff ratios from retention ratios: runoff = 1 - retention.

    Args:
        retention_array (numpy.ndarray): array of stormwater retention ratios.
            It is assumed that the nodata value is the global FLOAT_NODATA.

    Returns:
        numpy.ndarray of stormwater runoff ratios
    """
    runoff_array = numpy.full(retention_array.shape, FLOAT_NODATA,
                              dtype=numpy.float32)
    valid_mask = ~utils.array_equals_nodata(retention_array, FLOAT_NODATA)
    runoff_array[valid_mask] = 1 - retention_array[valid_mask]
    return runoff_array


def pollutant_load_op(lulc_array, lulc_nodata, volume_array, sorted_lucodes,
                      emc_array):
    """Calculate pollutant loads from EMC and stormwater volumes.

    Used for both actual pollutant load (where `volume_array` is the runoff
    volume) and avoided pollutant load (where `volume_array` is the
    retention volume).

    Args:
        lulc_array (numpy.ndarray): 2D array of LULC codes
        lulc_nodata (int): nodata value for the LULC array
        volume_array (numpy.ndarray): 2D array of stormwater volumes, with the
            same shape as ``lulc_array``. It is assumed that the volume nodata
            value is the global FLOAT_NODATA.
        sorted_lucodes (numpy.ndarray): 1D array of the LULC codes in order
            from smallest to largest
        emc_array (numpy.ndarray): 1D array of pollutant EMC values for each
            lucode. ``emc_array[i]`` is the EMC for the LULC class at
            ``sorted_lucodes[i]``.

    Returns:
        2D numpy.ndarray with the same shape as ``lulc_array``.
        Each value is the avoided pollutant load on that pixel in kg/yr,
        or FLOAT_NODATA if any of the inputs have nodata on that pixel.
    """
    load_array = numpy.full(
        lulc_array.shape, FLOAT_NODATA, dtype=numpy.float32)
    valid_mask = ~utils.array_equals_nodata(volume_array, FLOAT_NODATA)
    if lulc_nodata is not None:
        valid_mask &= ~utils.array_equals_nodata(lulc_array, lulc_nodata)

    # bin each value in the LULC array such that
    # lulc_array[i,j] == sorted_lucodes[lulc_index[i,j]]. thus,
    # emc_array[lulc_index[i,j]] is the EMC for the lucode at lulc_array[i,j]
    lulc_index = numpy.digitize(lulc_array, sorted_lucodes, right=True)
    # need to mask out nodata pixels from lulc_index before indexing emc_array
    # otherwise we get an IndexError when the nodata value is > all LULC values
    valid_lulc_index = lulc_index[valid_mask]
    # EMC for pollutant (mg/L) * 1000 (L/m^3) * 0.000001 (kg/mg) *
    # retention (m^3/yr) = pollutant load (kg/yr)
    load_array[valid_mask] = (emc_array[valid_lulc_index] *
                              0.001 * volume_array[valid_mask])
    return load_array


def retention_value_op(retention_volume_array, replacement_cost):
    """Multiply retention volumes by the retention replacement cost.

    Args:
        retention_volume_array (numpy.ndarray): 2D array of retention volumes.
            Assumes that the retention volume nodata value is the global
            FLOAT_NODATA.
        replacement_cost (float): Replacement cost per cubic meter of water

    Returns:
        numpy.ndarray of retention values with the same dimensions as the input
    """
    value_array = numpy.full(retention_volume_array.shape, FLOAT_NODATA,
                             dtype=numpy.float32)
    valid_mask = ~utils.array_equals_nodata(
        retention_volume_array, FLOAT_NODATA)

    # retention (m^3/yr) * replacement cost ($/m^3) = retention value ($/yr)
    value_array[valid_mask] = (
        retention_volume_array[valid_mask] * replacement_cost)
    return value_array


def adjust_op(ratio_array, avg_ratio_array, near_connected_lulc_array,
              near_road_array):
    """Apply the retention ratio adjustment algorithm to an array of ratios.

    This is meant to be used with raster_calculator. Assumes that the nodata
    value for all four input arrays is the global FLOAT_NODATA.

    Args:
        ratio_array (numpy.ndarray): 2D array of stormwater retention ratios
        avg_ratio_array (numpy.ndarray): 2D array of averaged ratios
        near_connected_lulc_array (numpy.ndarray): 2D boolean array where 1
            means this pixel is near a directly-connected LULC area
        near_road_array (numpy.ndarray): 2D boolean array where 1
            means this pixel is near a road centerline

    Returns:
        2D numpy array of adjusted retention ratios. Has the same shape as
        ``retention_ratio_array``.
    """
    adjusted_ratio_array = numpy.full(ratio_array.shape, FLOAT_NODATA,
                                      dtype=numpy.float32)
    adjustment_factor_array = numpy.full(ratio_array.shape, FLOAT_NODATA,
                                         dtype=numpy.float32)
    valid_mask = (
        ~utils.array_equals_nodata(ratio_array, FLOAT_NODATA) &
        ~utils.array_equals_nodata(avg_ratio_array, FLOAT_NODATA) &
        (near_connected_lulc_array != UINT8_NODATA) &
        (near_road_array != UINT8_NODATA))

    # adjustment factor:
    # - 0 if any of the nearby pixels are impervious/connected;
    # - average of nearby pixels, otherwise
    is_not_connected = ~(
        near_connected_lulc_array[valid_mask] |
        near_road_array[valid_mask]).astype(bool)
    adjustment_factor_array[valid_mask] = (avg_ratio_array[valid_mask] *
                                           is_not_connected)

    adjustment_factor_array[valid_mask] = (
        avg_ratio_array[valid_mask] * ~(
            near_connected_lulc_array[valid_mask] |
            near_road_array[valid_mask]
        ).astype(bool))

    # equation 2-4: Radj_ij = R_ij + (1 - R_ij) * C_ij
    adjusted_ratio_array[valid_mask] = (
        ratio_array[valid_mask] +
        (1 - ratio_array[valid_mask]) * adjustment_factor_array[valid_mask])
    return adjusted_ratio_array


def aggregate_results(base_aggregate_areas_path, target_vector_path, srs_wkt,
                      aggregations):
    """Aggregate outputs into regions of interest.

    Args:
        base_aggregate_areas_path (str): path to vector of polygon(s) to
            aggregate over. This is the original input.
        target_vector_path (str): path to write out the results. This will be a
            copy of the base vector with added fields, reprojected to the
            target WKT and saved in geopackage format.
        srs_wkt (str): a Well-Known Text representation of the target spatial
            reference. The base vector is reprojected to this spatial reference
            before aggregating the rasters over it.
        aggregations (list[tuple(str,str,str)]): list of tuples describing the
            datasets to aggregate. Each tuple has 3 items. The first is the
            path to a raster to aggregate. The second is the field name for
            this aggregated data in the output vector. The third is either
            'mean' or 'sum' indicating the aggregation to perform.

    Returns:
        None
    """
    pygeoprocessing.reproject_vector(base_aggregate_areas_path, srs_wkt,
                                     target_vector_path, driver_name='GPKG')
    aggregate_vector = gdal.OpenEx(target_vector_path, gdal.GA_Update)
    aggregate_layer = aggregate_vector.GetLayer()

    for raster_path, field_id, aggregation_op in aggregations:
        # aggregate the raster by the vector region(s)
        aggregate_stats = pygeoprocessing.zonal_statistics(
            (raster_path, 1), target_vector_path)

        # set up the field to hold the aggregate data
        aggregate_field = ogr.FieldDefn(field_id, ogr.OFTReal)
        aggregate_field.SetWidth(24)
        aggregate_field.SetPrecision(11)
        aggregate_layer.CreateField(aggregate_field)
        aggregate_layer.ResetReading()

        # save the aggregate data to the field for each feature
        for feature in aggregate_layer:
            feature_id = feature.GetFID()
            if aggregation_op == 'mean':
                pixel_count = aggregate_stats[feature_id]['count']
                try:
                    value = (aggregate_stats[feature_id]['sum'] / pixel_count)
                except ZeroDivisionError:
                    LOGGER.warning(
                        f'Polygon {feature_id} does not overlap {raster_path}')
                    value = 0.0
            elif aggregation_op == 'sum':
                value = aggregate_stats[feature_id]['sum']
            feature.SetField(field_id, float(value))
            aggregate_layer.SetFeature(feature)

    # save the aggregate vector layer and clean up references
    aggregate_layer.SyncToDisk()
    aggregate_layer = None
    gdal.Dataset.__swig_destroy__(aggregate_vector)
    aggregate_vector = None


def is_near(input_path, radius, distance_path, out_path):
    """Make binary raster of which pixels are within a radius of a '1' pixel.

    Args:
        input_path (str): path to a binary raster where '1' pixels are what
            we're measuring distance to, in this case roads/connected areas
        radius (float): distance in pixels which is considered "near".
            pixels this distance or less from a '1' pixel are marked '1' in
            the output. Distances are measured centerpoint to centerpoint.
        distance_path (str): path to write out the raster of distances
        out_path (str): path to write out the final thresholded raster.
            Pixels marked '1' are near to a '1' pixel in the input, pixels
            marked '0' are not.

    Returns:
        None
    """
    # Calculate the distance from each pixel to the nearest '1' pixel
    pygeoprocessing.distance_transform_edt(
        (input_path, 1),
        distance_path)

    def lte_threshold_op(array, threshold):
        """Binary array of elements less than or equal to the threshold."""
        # no need to mask nodata because distance_transform_edt doesn't
        # output any nodata pixels
        return array <= threshold

    # Threshold that to a binary array so '1' means it's within the radius
    pygeoprocessing.raster_calculator(
        [(distance_path, 1), (radius, 'raw')],
        lte_threshold_op,
        out_path,
        gdal.GDT_Byte,
        UINT8_NODATA)


def make_search_kernel(raster_path, radius):
    """Make a search kernel for a raster that marks pixels within a radius.

    Args:
        raster_path (str): path to a raster to make kernel for. It is assumed
            that the raster has square pixels.
        radius (float): distance around each pixel's centerpoint to search
            in raster coordinate system units

    Returns:
        2D boolean numpy.ndarray. '1' pixels are within ``radius`` of the
        center pixel, measured centerpoint-to-centerpoint. '0' pixels are
        outside the radius. The array dimensions are as small as possible
        while still including the entire radius.
    """
    raster_info = pygeoprocessing.get_raster_info(raster_path)
    pixel_radius = radius / abs(raster_info['pixel_size'][0])
    pixel_margin = math.floor(pixel_radius)
    # the search kernel is just large enough to contain all pixels that
    # *could* be within the radius of the center pixel
    search_kernel_shape = tuple([pixel_margin * 2 + 1] * 2)
    # arrays of the column index and row index of each pixel
    col_indices, row_indices = numpy.indices(search_kernel_shape)
    # adjust them so that (0, 0) is the center pixel
    col_indices -= pixel_margin
    row_indices -= pixel_margin
    # hypotenuse_i = sqrt(col_indices_i**2 + row_indices_i**2) for each pixel i
    hypotenuse = numpy.hypot(col_indices, row_indices)
    # boolean kernel where 1=pixel centerpoint is within the radius of the
    # center pixel's centerpoint
    search_kernel = numpy.array(hypotenuse <= pixel_radius, dtype=numpy.uint8)
    LOGGER.debug(
        f'Search kernel for {raster_path} with radius {radius}:'
        f'\n{search_kernel}')
    return search_kernel


def raster_average(raster_path, radius, kernel_path, out_path):
    """Average pixel values within a radius.

    Make a search kernel where a pixel has '1' if its centerpoint is within
    the radius of the center pixel's centerpoint.
    For each pixel in a raster, center the search kernel on top of it. Then
    its "neighborhood" includes all the pixels that are below a '1' in the
    search kernel. Add up the neighborhood pixel values and divide by how
    many there are.

    This accounts for edge pixels and nodata pixels. For instance, if the
    kernel covers a 3x3 pixel area centered on each pixel, most pixels will
    have 9 valid pixels in their neighborhood, most edge pixels will have 6,
    and most corner pixels will have 4. Edge and nodata pixels in the
    neighborhood don't count towards the total (denominator in the average).

    Args:
        raster_path (str): path to the raster file to average
        radius (float): distance to average around each pixel's centerpoint in
            raster coordinate system units
        kernel_path (str): path to write out the search kernel raster, an
            intermediate output required by pygeoprocessing.convolve_2d
        out_path (str): path to write out the averaged raster output

    Returns:
        None
    """
    search_kernel = make_search_kernel(raster_path, radius)

    srs = osr.SpatialReference()
    srs.ImportFromEPSG(3857)
    projection_wkt = srs.ExportToWkt()
    pygeoprocessing.numpy_array_to_raster(
        # float32 here to avoid pygeoprocessing bug issue #180
        search_kernel.astype(numpy.float32),
        FLOAT_NODATA,
        (20, -20),
        (0, 0),
        projection_wkt,
        kernel_path)

    # convolve the signal (input raster) with the kernel and normalize
    # this is equivalent to taking an average of each pixel's neighborhood
    pygeoprocessing.convolve_2d(
        (raster_path, 1),
        (kernel_path, 1),
        out_path,
        # pixels with nodata or off the edge of the raster won't count towards
        # the sum or the number of values to normalize by
        ignore_nodata_and_edges=True,
        # divide by number of valid pixels in the kernel (averaging)
        normalize_kernel=True,
        # output will have nodata where ratio_path has nodata
        mask_nodata=True,
        target_datatype=gdal.GDT_Float32,
        target_nodata=FLOAT_NODATA)


@ validation.invest_validator
def validate(args, limit_to=None):
    """Validate args to ensure they conform to `execute`'s contract.

    Args:
        args (dict): dictionary of key(str)/value pairs where keys and
            values are specified in `execute` docstring.
        limit_to (str): (optional) if not None indicates that validation
            should only occur on the args[limit_to] value. The intent that
            individual key validation could be significantly less expensive
            than validating the entire `args` dictionary.

    Returns:
        list of ([invalid key_a, invalid_keyb, ...], 'warning/error message')
            tuples. Where an entry indicates that the invalid keys caused
            the error message in the second part of the tuple. This should
            be an empty list if validation succeeds.
    """
    return validation.validate(args, ARGS_SPEC['args'],
                               ARGS_SPEC['args_with_spatial_overlap'])
